# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring,unsubscriptable-object,protected-access
import unittest
from unittest.mock import Mock, patch

import numpy as np
from AnyQt.QtCore import Qt, QModelIndex, QItemSelectionModel
from AnyQt.QtTest import QSignalSpy

from orangewidget.tests.base import GuiTest
from orangewidget.utils.itemmodels import SeparatedListDelegate

from Orange.data import Table, DiscreteVariable, ContinuousVariable, Domain
from Orange.widgets.data.owcontinuize import OWContinuize, DefaultKey, \
    ContinuousOptions, Normalize, Continuize, DiscreteOptions, \
    ContDomainModel, DefaultContModel, ListViewSearch, DefaultId
from Orange.widgets.tests.base import WidgetTest


class TestOWContinuize(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWContinuize)

    def test_empty_data(self):
        """No crash on empty data"""
        data = Table("iris")
        widget = self.widget

        self.send_signal(self.widget.Inputs.data, data)
        widget.commit.now()

        self.send_signal(self.widget.Inputs.data,
                         Table.from_domain(data.domain))
        widget.commit.now()
        self.assertIsNone(self.get_output(self.widget.Outputs.data))

        self.send_signal(self.widget.Inputs.data, None)
        widget.commit.now()
        self.assertIsNone(self.get_output(self.widget.Outputs.data))

    def test_continuous(self):
        table = Table("housing")
        self.send_signal(self.widget.Inputs.data, table)
        self.widget.commit.now()

    def test_one_column_equal_values(self):
        table = Table("iris")
        table = table[:, 1].copy()
        with table.unlocked():
            table[:] = 42.0
        self.send_signal(self.widget.Inputs.data, table)
        # Normalize.NormalizeBySD
        self.widget.continuous_treatment = 2
        self.widget.commit.now()

    def test_one_column_nan_values_normalize_sd(self):
        table = Table("iris")
        with table.unlocked():
            table[:, 2] = np.NaN
        self.send_signal(self.widget.Inputs.data, table)
        # Normalize.NormalizeBySD
        self.widget.continuous_treatment = 2
        self.widget.commit.now()

        table = Table("iris")
        with table.unlocked():
            table[1, 2] = np.NaN
        self.send_signal(self.widget.Inputs.data, table)
        self.widget.commit.now()

    def test_one_column_nan_values_normalize_span(self):
        table = Table("iris")
        with table.unlocked():
            table[:, 2] = np.NaN
        self.send_signal(self.widget.Inputs.data, table)
        # Normalize.NormalizeBySpan
        self.widget.continuous_treatment = 1
        self.widget.commit.now()

        table = Table("iris")
        with table.unlocked():
            table[1, 2] = np.NaN
        self.send_signal(self.widget.Inputs.data, table)
        self.widget.commit.now()

    def test_commit_calls_prepare_output(self):
        # This test ensures that commit returns the result of _prepare_output,
        # so further tests can just check the latter. If this is changed, the
        # test will fail, which is OK - test can be removed, but other tests
        # then have to check the output and not just _prepare_output.
        out = object()
        self.widget._prepare_output = lambda: out
        self.widget.Outputs.data.send = Mock()
        self.widget.commit.now()
        self.widget.Outputs.data.send.assert_called_with(out)

    def test_check_unsuppoerted_sparse_continuous(self):
        # This test checks response at two points:
        # - when scaling sparse data with a method that does not support it,
        #   the wiget must show an error and output nothing
        # - the above is tested via method _unsupported_sparse, so we also
        #   directly check this method
        w = self.widget
        hints = w.cont_var_hints
        iris = Table("iris")
        iris = iris.transform(Domain(iris.domain[:2],
                                     iris.domain.class_var,
                                     iris.domain.attributes[2:]))
        sparse_iris = iris.to_sparse(sparse_class=True, sparse_metas=True)

        for attr in (iris.domain.attributes[0], iris.domain.metas[0]):
            for key in (DefaultKey, attr.name):
                hints[DefaultKey] = Normalize.Leave
                for hints[key], desc in ContinuousOptions.items():
                    if desc.id_ == Normalize.Default:
                        continue
                    msg = f"at {attr} = {desc.label}, " \
                          + ("default" if key is DefaultKey else key)

                    # input dense
                    self.send_signal(w.Inputs.data, iris)
                    self.assertFalse(w._unsupported_sparse(), msg)
                    self.assertFalse(w.Error.unsupported_sparse.is_shown(), msg)
                    self.assertIsNotNone(self.get_output(w.Outputs.data), msg)

                    # input sparse
                    self.send_signal(w.Inputs.data, sparse_iris)
                    self.assertIsNot(w._unsupported_sparse(),
                                     desc.supports_sparse, msg)
                    self.assertIsNot(w.Error.unsupported_sparse.is_shown(),
                                     desc.supports_sparse, msg)
                    if desc.supports_sparse:
                        self.assertIsNotNone(self.get_output(w.Outputs.data),
                                             msg)
                    else:
                        self.assertIsNone(self.get_output(w.Outputs.data),
                                          msg)
                        self.send_signal(w.Inputs.data, None)
                        self.assertFalse(w.Error.unsupported_sparse.is_shown(),
                                         msg)
                del hints[key]

    def test_check_unsuppoerted_sparse_discrete(self):
        # This test checks response at two points:
        # - when scaling sparse data with a method that does not support it,
        #   the wiget must show an error and output nothing
        # - the above is tested via method _unsupported_sparse, so we also
        #   directly check this method
        w = self.widget
        hints = w.disc_var_hints
        zoo = Table("zoo")
        zoo = zoo.transform(Domain(zoo.domain.attributes[:2],
                                   None,
                                   zoo.domain.attributes[2:]))
        sparse_zoo = zoo.to_sparse(sparse_metas=True)

        # input dense
        for attr in (zoo.domain[0], zoo.domain.metas[0]):
            for key in (DefaultKey, attr.name):
                hints[DefaultKey] = Continuize.Leave
                for hints[key], desc in DiscreteOptions.items():
                    if desc.id_ == Continuize.Default:
                        continue
                    msg = f"at {key} = {desc.label}, " \
                          + ("default" if key is DefaultKey else key)

                    self.send_signal(w.Inputs.data, zoo)
                    self.assertFalse(w._unsupported_sparse(), msg)
                    self.assertFalse(w.Error.unsupported_sparse.is_shown(), msg)
                    self.assertIsNotNone(self.get_output(w.Outputs.data), msg)

                    self.send_signal(w.Inputs.data, sparse_zoo)
                    self.assertIsNot(w._unsupported_sparse(),
                                     desc.supports_sparse, msg)
                    self.assertIsNot(w.Error.unsupported_sparse.is_shown(),
                                     desc.supports_sparse, msg)
                    if desc.supports_sparse:
                        self.assertIsNotNone(self.get_output(w.Outputs.data), msg)
                    else:
                        self.assertIsNone(self.get_output(w.Outputs.data), msg)
                        self.send_signal(w.Inputs.data, None)
                        self.assertFalse(w.Error.unsupported_sparse.is_shown(), msg)
            del hints[key]

    def test_update_cont_radio_buttons(self):
        w = self.widget
        w.disc_var_hints[DefaultKey] = Continuize.AsOrdinal
        w.disc_var_hints["chest pain"] \
            = w.disc_var_hints["rest ECG"] \
            = Continuize.Remove
        w.disc_var_hints["exerc ind ang"] = Continuize.FirstAsBase

        w.cont_var_hints[DefaultKey] = Normalize.Center
        w.cont_var_hints["cholesterol"] = Normalize.Scale

        self.send_signal(w.Inputs.data, Table("heart_disease"))

        dview = w.disc_view
        dmod = dview.model()
        dselmod = dview.selectionModel()
        dgroup = w.disc_group

        with patch.object(w, "_update_radios") as upd:
            w._on_var_selection_changed(dview)
            upd.assert_not_called()

        dselmod.select(dmod.index(1, 0),
                       QItemSelectionModel.ClearAndSelect)  # chest_pain
        self.assertEqual(dgroup.checkedId(), Continuize.Remove)
        self.assertTrue(dgroup.button(99).isEnabled())

        dselmod.select(dmod.index(2, 0),
                       QItemSelectionModel.ClearAndSelect)  # blood sugar
        self.assertEqual(dgroup.checkedId(), Continuize.Default)

        dselmod.select(dmod.index(3, 0),
                       QItemSelectionModel.ClearAndSelect)  # rest ECG
        self.assertEqual(dgroup.checkedId(), Continuize.Remove)

        dselmod.select(dmod.index(4, 0),
                       QItemSelectionModel.ClearAndSelect)  # exerc ind ang
        self.assertEqual(dgroup.checkedId(), Continuize.FirstAsBase)

        dselmod.select(dmod.index(3, 0),
                       QItemSelectionModel.Select)  # read ECG and exerc ind ang
        self.assertEqual(dgroup.checkedId(), -1)

        dview.select_default()
        self.assertEqual(dgroup.checkedId(), Continuize.AsOrdinal)
        self.assertFalse(dgroup.button(99).isEnabled())

        cview = w.cont_view
        cmod = cview.model()
        cselmod = cview.selectionModel()
        cgroup = w.cont_group

        cselmod.select(cmod.index(2, 0),
                       QItemSelectionModel.ClearAndSelect)  # cholesterol
        self.assertEqual(cgroup.checkedId(), Normalize.Scale)
        self.assertEqual(dgroup.checkedId(), Continuize.AsOrdinal)
        self.assertTrue(cgroup.button(99).isEnabled())

        cview.select_default()
        self.assertEqual(cgroup.checkedId(), Normalize.Center)
        self.assertEqual(dgroup.checkedId(), Continuize.AsOrdinal)
        self.assertFalse(cgroup.button(99).isEnabled())

        w._uncheck_all_buttons(cgroup)
        self.assertEqual(cgroup.checkedId(), -1)
        self.assertEqual(dgroup.checkedId(), Continuize.AsOrdinal)

        w._uncheck_all_buttons(dgroup)
        self.assertEqual(dgroup.checkedId(), -1)

    def test_update_disc_radio_buttons_mixed(self):
        def select(xs):
            dselmod.select(dmod.index(xs[0], 0),
                           QItemSelectionModel.ClearAndSelect)
            for x in xs[1:]:
                dselmod.select(dmod.index(x, 0),
                               QItemSelectionModel.Select)

        w = self.widget
        dview = w.disc_view
        dmod = dview.model()
        dselmod = dview.selectionModel()
        dgroup = w.disc_group

        domain = Domain(
            [DiscreteVariable(x, values=["0", "1"]) for x in "abc"],
            DiscreteVariable("d", values=["0", "1"]),
            [DiscreteVariable(x, values=["0", "1"]) for x in "efg"])
        data = Table.from_list(domain, [[1] * 7] * 2)

        w.disc_var_hints = {DefaultKey: Continuize.FirstAsBase}
        self.send_signal(w.Inputs.data, data)
        select([1, 4, 8])
        self.assertEqual(dgroup.checkedId(), -1)
        select([1])
        self.assertEqual(dgroup.checkedId(), DefaultId)
        select([4, 8])
        self.assertEqual(dgroup.checkedId(), Continuize.Leave)

        w.disc_var_hints = {DefaultKey: Continuize.FirstAsBase,
                            "b": Continuize.Leave}
        self.send_signal(w.Inputs.data, data)
        select([1, 4, 8])
        self.assertEqual(dgroup.checkedId(), Continuize.Leave)

        w.disc_var_hints = {DefaultKey: Continuize.FirstAsBase,
                            "e": DefaultId, "d": DefaultId}
        self.send_signal(w.Inputs.data, data)
        select([1, 4, 8])
        self.assertEqual(dgroup.checkedId(), DefaultId)

    def test_update_cont_radio_buttons_mixed(self):
        def select(xs):
            cselmod.select(cmod.index(xs[0], 0),
                           QItemSelectionModel.ClearAndSelect)
            for x in xs[1:]:
                cselmod.select(cmod.index(x, 0),
                               QItemSelectionModel.Select)

        w = self.widget
        cview = w.cont_view
        cmod = cview.model()
        cselmod = cview.selectionModel()
        cgroup = w.cont_group

        domain = Domain([ContinuousVariable(x) for x in "abc"],
                        ContinuousVariable("d"),
                        [ContinuousVariable(x) for x in "efg"])
        data = Table.from_list(domain, [[1] * 7] * 2)

        w.cont_var_hints = {DefaultKey: Normalize.Center}
        self.send_signal(w.Inputs.data, data)
        select([1, 4, 8])
        self.assertEqual(cgroup.checkedId(), -1)
        select([1])
        self.assertEqual(cgroup.checkedId(), DefaultId)
        select([4, 8])
        self.assertEqual(cgroup.checkedId(), Normalize.Leave)

        w.cont_var_hints = {DefaultKey: Normalize.Center,
                            "b": Normalize.Leave}
        self.send_signal(w.Inputs.data, data)
        select([1, 4, 8])
        self.assertEqual(cgroup.checkedId(), Normalize.Leave)

        w.cont_var_hints = {DefaultKey: Normalize.Center,
                            "e": DefaultId, "d": DefaultId}
        self.send_signal(w.Inputs.data, data)
        select([1, 4, 8])
        self.assertEqual(cgroup.checkedId(), DefaultId)

    def test_set_hints_on_new_data(self):
        w = self.widget
        domain = Domain([ContinuousVariable(c) for c in "abc"] +
                        [DiscreteVariable("m", values=tuple("xy"))],
                        ContinuousVariable("d"),
                        [ContinuousVariable(c) for c in "ef"])
        data = Table.from_list(domain, [[0] * 6])

        w.cont_var_hints["b"] = Normalize.Leave
        w.cont_var_hints["f"] = Normalize.Normalize11
        w.cont_var_hints["x"] = Normalize.Normalize11

        self.send_signal(w.Inputs.data, None)
        self.send_signal(w.Inputs.data, data)

        model = w.cont_view.model()
        self.assertEqual(model.index(0, 0).data(model.HintRole),
                         ("preset", False))
        self.assertEqual(model.index(1, 0).data(model.HintRole),
                         (ContinuousOptions[Normalize.Leave].short_desc, True))
        self.assertEqual(model.index(5, 0).data(model.HintRole),
                         (ContinuousOptions[Normalize.Normalize11].short_desc, True))
        self.assertNotIn("x", w.cont_var_hints)

    def test_reset_hints(self):
        w = self.widget
        domain = Domain([ContinuousVariable(c) for c in "abc"] +
                        [DiscreteVariable("m", values=tuple("xy"))],
                        ContinuousVariable("d"),
                        [ContinuousVariable(c) for c in "ef"])
        data = Table.from_list(domain, [[0] * 6])

        w.cont_var_hints[DefaultKey] = Normalize.Center
        w.cont_var_hints["b"] = Normalize.Leave
        w.cont_var_hints["f"] = Normalize.Normalize11
        w.cont_var_hints["x"] = Normalize.Normalize11
        w.disc_var_hints[DefaultKey] = Continuize.Indicators
        w.cont_var_hints["m"] = Continuize.Remove

        self.send_signal(w.Inputs.data, data)
        w._on_reset_hints()

        self.assertEqual(w.cont_var_hints[DefaultKey], Normalize.Leave)
        self.assertEqual(w.disc_var_hints[DefaultKey], Continuize.FirstAsBase)

    def test_change_hints_disc(self):
        w = self.widget
        w.disc_var_hints[DefaultKey] = Continuize.AsOrdinal
        w.disc_var_hints["chest pain"] \
            = w.disc_var_hints["rest ECG"] \
            = Continuize.Remove
        w.disc_var_hints["exerc ind ang"] = Continuize.FirstAsBase

        dview = w.disc_view
        dmod = dview.model()
        dselmod = dview.selectionModel()
        dgroup = w.disc_group

        self.send_signal(w.Inputs.data, Table("heart_disease"))
        self.assertEqual(
            dmod.index(3, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Remove].short_desc, True))

        dselmod.select(dmod.index(1, 0),
                       QItemSelectionModel.ClearAndSelect)  # chest pain
        dselmod.select(dmod.index(4, 0),
                       QItemSelectionModel.Select)  # exerc ind ang
        dgroup.button(Continuize.AsOrdinal).setChecked(True)
        dgroup.idClicked.emit(Continuize.AsOrdinal)

        self.assertFalse("gender" in w.disc_var_hints)
        self.assertEqual(w.disc_var_hints["chest pain"], Continuize.AsOrdinal)
        self.assertEqual(w.disc_var_hints["exerc ind ang"], Continuize.AsOrdinal)
        self.assertEqual(w.disc_var_hints["rest ECG"], Continuize.Remove)

        dselmod.select(dmod.index(1, 0),
                       QItemSelectionModel.ClearAndSelect)  # chest pain
        dselmod.select(dmod.index(0, 0),
                       QItemSelectionModel.Select)  # gender
        dgroup.button(99).setChecked(True)
        dgroup.idClicked.emit(99)
        self.assertFalse("chest pain" in w.disc_var_hints)
        self.assertFalse("gender" in w.disc_var_hints)
        self.assertEqual(w.disc_var_hints["rest ECG"], Continuize.Remove)

        self.assertEqual(dmod.index(0, 0).data(dmod.HintRole),
                         ("preset", False))
        self.assertEqual(
            dmod.index(3, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Remove].short_desc, True))

        dview.select_default()
        dgroup.button(Continuize.AsOrdinal).setChecked(True)
        dgroup.idClicked.emit(Continuize.AsOrdinal)
        self.assertEqual(w.disc_var_hints[DefaultKey], Continuize.AsOrdinal)

    def test_change_hints_disc_class_meta(self):
        w = self.widget
        dview = w.disc_view
        dmod = dview.model()
        dselmod = dview.selectionModel()
        dgroup = w.disc_group

        domain = Domain([DiscreteVariable(x, values=["0", "1"]) for x in "abc"],
                        DiscreteVariable("d", values=["0", "1"]),
                        [DiscreteVariable(x, values=["0", "1"]) for x in "efg"])
        data = Table.from_list(domain, [[1] * 7] * 2)
        self.send_signal(w.Inputs.data, data)

        dselmod.select(dmod.index(1, 0),
                       QItemSelectionModel.ClearAndSelect)  # attribute b
        dselmod.select(dmod.index(4, 0),
                       QItemSelectionModel.Select)  # meta e
        dselmod.select(dmod.index(8, 0),
                       QItemSelectionModel.Select)  # class d
        dgroup.button(Continuize.Remove).setChecked(True)
        dgroup.idClicked.emit(Continuize.Remove)
        self.assertEqual(w.disc_var_hints["b"], Continuize.Remove)
        self.assertEqual(w.disc_var_hints["e"], Continuize.Remove)
        self.assertEqual(w.disc_var_hints["d"], Continuize.Remove)
        self.assertEqual(
            dmod.index(1, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Remove].short_desc, True))
        self.assertEqual(
            dmod.index(4, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Remove].short_desc, True))
        self.assertEqual(
            dmod.index(8, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Remove].short_desc, True))

        dgroup.button(DefaultId).setChecked(True)
        dgroup.idClicked.emit(DefaultId)
        self.assertNotIn("b", w.disc_var_hints)
        self.assertEqual(w.disc_var_hints["e"], DefaultId)
        self.assertEqual(w.disc_var_hints["d"], DefaultId)
        self.assertEqual(
            dmod.index(1, 0).data(dmod.HintRole),
            (DiscreteOptions[DefaultId].short_desc, False))
        self.assertEqual(
            dmod.index(4, 0).data(dmod.HintRole),
            (DiscreteOptions[DefaultId].short_desc, True))
        self.assertEqual(
            dmod.index(8, 0).data(dmod.HintRole),
            (DiscreteOptions[DefaultId].short_desc, True))

        dgroup.button(Continuize.Leave).setChecked(True)
        dgroup.idClicked.emit(Continuize.Leave)
        self.assertEqual(w.disc_var_hints["b"], Continuize.Leave)
        self.assertNotIn("e", w.disc_var_hints)
        self.assertNotIn("d", w.disc_var_hints)
        self.assertEqual(
            dmod.index(1, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Leave].short_desc, True))
        self.assertEqual(
            dmod.index(4, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Leave].short_desc, False))
        self.assertEqual(
            dmod.index(8, 0).data(dmod.HintRole),
            (DiscreteOptions[Continuize.Leave].short_desc, False))

    def test_change_hints_cont(self):
        w = self.widget
        w.cont_var_hints[DefaultKey] = Normalize.Center
        w.cont_var_hints["cholesterol"] = Normalize.Scale

        self.send_signal(w.Inputs.data, Table("heart_disease"))

        cview = w.cont_view
        cmod = cview.model()
        cselmod = cview.selectionModel()
        cgroup = w.cont_group

        cselmod.select(cmod.index(2, 0),
                       QItemSelectionModel.ClearAndSelect)  # cholesterol
        cselmod.select(cmod.index(3, 0),
                       QItemSelectionModel.Select)  # max HR
        cgroup.button(Normalize.Normalize11).setChecked(True)
        cgroup.idClicked.emit(Normalize.Normalize11)

        self.assertFalse("age" in w.cont_var_hints)
        self.assertEqual(w.cont_var_hints["cholesterol"], Normalize.Normalize11)
        self.assertEqual(w.cont_var_hints["max HR"], Normalize.Normalize11)

        cselmod.select(cmod.index(2, 0),
                       QItemSelectionModel.ClearAndSelect)  # cholesterol
        cselmod.select(cmod.index(0, 0),
                       QItemSelectionModel.Select)  # age
        cgroup.button(99).setChecked(True)
        cgroup.idClicked.emit(99)
        self.assertFalse("age" in w.cont_var_hints)
        self.assertFalse("cholesterol" in w.cont_var_hints)
        self.assertEqual(w.cont_var_hints["max HR"], Normalize.Normalize11)

        self.assertEqual(cmod.index(0, 0).data(cmod.HintRole),
                         ("preset", False))
        self.assertEqual(
            cmod.index(3, 0).data(cmod.HintRole),
            (ContinuousOptions[Normalize.Normalize11].short_desc, True))

    def test_change_hints_cont_class_meta(self):
        w = self.widget
        cview = w.cont_view
        cmod = cview.model()
        cselmod = cview.selectionModel()
        cgroup = w.cont_group

        domain = Domain([ContinuousVariable(x) for x in "abc"],
                        ContinuousVariable("d"),
                        [ContinuousVariable(x) for x in "efg"])
        data = Table.from_list(domain, [[1] * 7] * 2)
        self.send_signal(w.Inputs.data, data)

        cselmod.select(cmod.index(1, 0),
                       QItemSelectionModel.ClearAndSelect)  # attribute b
        cselmod.select(cmod.index(4, 0),
                       QItemSelectionModel.Select)  # meta e
        cselmod.select(cmod.index(8, 0),
                       QItemSelectionModel.Select)  # class d
        cgroup.button(Normalize.Center).setChecked(True)
        cgroup.idClicked.emit(Normalize.Center)
        self.assertEqual(w.cont_var_hints["b"], Normalize.Center)
        self.assertEqual(w.cont_var_hints["e"], Normalize.Center)
        self.assertEqual(w.cont_var_hints["d"], Normalize.Center)
        self.assertEqual(
            cmod.index(1, 0).data(cmod.HintRole),
            (ContinuousOptions[Normalize.Center].short_desc, True))
        self.assertEqual(
            cmod.index(4, 0).data(cmod.HintRole),
            (ContinuousOptions[Normalize.Center].short_desc, True))
        self.assertEqual(
            cmod.index(8, 0).data(cmod.HintRole),
            (ContinuousOptions[Normalize.Center].short_desc, True))

        cgroup.button(DefaultId).setChecked(True)
        cgroup.idClicked.emit(DefaultId)
        self.assertNotIn("b", w.cont_var_hints)
        self.assertEqual(w.cont_var_hints["e"], DefaultId)
        self.assertEqual(w.cont_var_hints["d"], DefaultId)
        self.assertEqual(
            cmod.index(1, 0).data(cmod.HintRole),
            (ContinuousOptions[DefaultId].short_desc, False))
        self.assertEqual(
            cmod.index(4, 0).data(cmod.HintRole),
            (ContinuousOptions[DefaultId].short_desc, True))
        self.assertEqual(
            cmod.index(8, 0).data(cmod.HintRole),
            (ContinuousOptions[DefaultId].short_desc, True))

        cgroup.button(Normalize.Leave).setChecked(True)
        cgroup.idClicked.emit(Normalize.Leave)
        self.assertEqual(w.cont_var_hints["b"], Normalize.Leave)
        self.assertNotIn("e", w.cont_var_hints)
        self.assertNotIn("d", w.cont_var_hints)
        self.assertEqual(
            cmod.index(1, 0).data(cmod.HintRole),
            (ContinuousOptions[Normalize.Leave].short_desc, True))
        self.assertEqual(
            cmod.index(4, 0).data(cmod.HintRole),
            (ContinuousOptions[Normalize.Leave].short_desc, False))
        self.assertEqual(
            cmod.index(8, 0).data(cmod.HintRole),
            (ContinuousOptions[Normalize.Leave].short_desc, False))

    def test_is_attr_and_default(self):
        w = self.widget
        a, b, d, e, f = (ContinuousVariable(x) for x in "abdef")
        c, g = (DiscreteVariable(x, values=["0", "1"]) for x in "cg")
        domain = Domain([a, b, c], d, [e, f, g])
        data = Table.from_list(domain, [[1] * 7] * 2)
        self.send_signal(w.Inputs.data, data)

        self.assertTrue(w.is_attr(a))
        self.assertTrue(w.is_attr(b))
        self.assertTrue(w.is_attr(c))
        self.assertFalse(w.is_attr(d))
        self.assertFalse(w.is_attr(e))
        self.assertFalse(w.is_attr(f))
        self.assertFalse(w.is_attr(g))

        self.assertEqual(w.default_for_var(a), DefaultId)
        self.assertEqual(w.default_for_var(c), DefaultId)
        self.assertEqual(w.default_for_var(d), Normalize.Leave)
        self.assertEqual(w.default_for_var(e), Normalize.Leave)
        self.assertEqual(w.default_for_var(g), Continuize.Leave)

    def test_hint_for_var(self):
        w = self.widget
        c1, c2, c3, c4, c5 = (ContinuousVariable(f"c{x}") for x in range(1, 6))
        d1, d2, d3, d4 = (DiscreteVariable(f"d{x}", values=["0", "1"]) for x in range(1, 5))
        domain = Domain([c1, c2, d1, d2], c5, [c3, c4, d3, d4])
        data = Table.from_list(domain, [[1] * 7] * 2)
        w.cont_var_hints = {
            DefaultKey: Normalize.Center,
            "c1": Normalize.Scale,
            "c3": Normalize.Standardize,
        }
        w.disc_var_hints = {
            DefaultKey: Continuize.FrequentAsBase,
            "d1": Continuize.Remove,
            "d3": Continuize.Indicators
        }
        self.send_signal(w.Inputs.data, data)

        self.assertEqual(w._hint_for_var(c1), Normalize.Scale)
        self.assertEqual(w._hint_for_var(c2), Normalize.Center)
        self.assertEqual(w._hint_for_var(c3), Normalize.Standardize)
        self.assertEqual(w._hint_for_var(c4), Normalize.Leave)

        self.assertEqual(w._hint_for_var(d1), Continuize.Remove)
        self.assertEqual(w._hint_for_var(d2), Continuize.FrequentAsBase)
        self.assertEqual(w._hint_for_var(d3), Continuize.Indicators)
        self.assertEqual(w._hint_for_var(d4), Continuize.Leave)

    def test_transformations(self):
        domain = Domain([DiscreteVariable(c, values="abc")
                         for c in ("default", "leave", "first", "frequent",
                                   "one-hot", "remove-if", "remove", "ordinal",
                                   "normordinal")],
                         DiscreteVariable("y", values="abc"),
                        [ContinuousVariable(c)
                         for c in ("cdefault", "cleave",
                                   "cstandardize", "ccenter", "cscale",
                                   "cnormalize11", "cnormalize01")]
                        )
        data = Table.from_list(domain,
                               [[x] * 17 for x in range(3)] + [[2] * 17])

        w = self.widget
        w.disc_var_hints = {
            var.name: id_
            for var, id_ in zip(domain.attributes, DiscreteOptions)
            if id_ != 99
        }
        w.disc_var_hints[DefaultKey] = Continuize.FrequentAsBase

        w.cont_var_hints = {
            var.name: id_
            for var, id_ in zip(domain.metas, ContinuousOptions)
            if id_ != 99
        }
        w.cont_var_hints[DefaultKey] = Normalize.Center

        self.send_signal(w.Inputs.data, data)
        outp = self.get_output(w.Outputs.data)

        np.testing.assert_almost_equal(
            outp.X,
            [[1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
             [0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0.5],
             [0, 0, 2, 0, 1, 0, 0, 0, 0, 1, 2, 1],
             [0, 0, 2, 0, 1, 0, 0, 0, 0, 1, 2, 1],
             ]
        )
        np.testing.assert_almost_equal(
            outp.Y,
            [0, 1, 2, 2]
        )
        np.testing.assert_almost_equal(
            outp.metas,
            [[0, 0, -1.50755672, -1.25, 0, -1, 0],
             [1, 1, -0.30151134, -0.25, 1.20604538, 0, 0.5],
             [2, 2, 0.90453403, 0.75, 2.41209076, 1, 1],
             [2, 2, 0.90453403, 0.75, 2.41209076, 1, 1],
             ]
        )

    def test_send_report(self):
        w = self.widget
        self.send_signal(w.Inputs.data, Table("heart_disease"))
        self.widget.send_report()

        w.disc_var_hints[DefaultKey] = Continuize.AsOrdinal
        w.disc_var_hints["chest pain"] \
            = w.disc_var_hints["rest ECG"] \
            = Continuize.Remove
        w.disc_var_hints["exerc ind ang"] = Continuize.FirstAsBase

        self.send_signal(w.Inputs.data, Table("heart_disease"))
        self.widget.send_report()

        w.cont_var_hints[DefaultKey] = Normalize.Center
        w.cont_var_hints["cholesterol"] = Normalize.Scale

        self.send_signal(w.Inputs.data, Table("heart_disease"))
        self.widget.send_report()

        w.continuize_class = True
        w.disc_var_hints[DefaultKey] = Continuize.AsOrdinal
        w.disc_var_hints["chest pain"] \
            = w.disc_var_hints["rest ECG"] \
            = Continuize.Remove
        w.disc_var_hints["exerc ind ang"] = Continuize.FirstAsBase

        w.cont_var_hints[DefaultKey] = Normalize.Center
        w.cont_var_hints["cholesterol"] = Normalize.Scale

        self.send_signal(w.Inputs.data, Table("heart_disease"))
        self.widget.send_report()

    def test_migrate_settings_to_v3(self):
        # why not?, pylint: disable=use-dict-literal
        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(continuous_treatment=0))
        self.assertEqual(widget.cont_var_hints[DefaultKey],
                         Normalize.Leave)

        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(continuous_treatment=1, zero_based=True))
        self.assertEqual(widget.cont_var_hints[DefaultKey],
                         Normalize.Normalize01)

        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(continuous_treatment=1, zero_based=False))
        self.assertEqual(widget.cont_var_hints[DefaultKey],
                         Normalize.Normalize11)

        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(continuous_treatment=2))
        self.assertEqual(widget.cont_var_hints[DefaultKey],
                         Normalize.Standardize)

        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(multinomial_treatment=2)
        )
        self.assertEqual(widget.disc_var_hints[DefaultKey],
                         Continuize.Indicators)

    def test_migrate_settings_to_v3_class_treatment(self):
        # why not?, pylint: disable=use-dict-literal
        domain = Domain([ContinuousVariable(c) for c in "abc"],
                        DiscreteVariable("y"))
        data = Table.from_list(domain, [[0] * 4] * 2)

        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(multinomial_treatment=4,
                                 class_treatment=3)
        )
        self.send_signal(widget.Inputs.data, data)
        self.assertEqual(widget.disc_var_hints["y"], Continuize.Indicators)
        self.assertEqual(widget.disc_var_hints[DefaultKey], Continuize.Remove)

        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(multinomial_treatment=4,
                                 class_treatment=0)
        )
        self.send_signal(widget.Inputs.data, data)
        self.assertNotIn("y", widget.disc_var_hints)
        self.assertEqual(widget.disc_var_hints[DefaultKey], 4)

        widget = self.create_widget(
            OWContinuize,
            stored_settings=dict(multinomial_treatment=Continuize.Remove)
        )
        self.send_signal(widget.Inputs.data, data)
        self.assertNotIn("y", widget.disc_var_hints)
        self.assertEqual(widget.disc_var_hints[DefaultKey], Continuize.Remove)


class TestModelsAndViews(GuiTest):
    def test_contmodel(self):
        domain = Domain([ContinuousVariable(c) for c in "abc"],
                        ContinuousVariable("y"))
        model = ContDomainModel(ContinuousVariable)
        model.set_domain(domain)

        ind = model.index(0, 0)
        self.assertEqual(ind.data()[0], "a")
        self.assertEqual(ind.data(model.FilterRole)[0], "a")
        self.assertIsNone(ind.data(Qt.ToolTipRole))

        ind = model.index(1, 0)
        model.setData(ind, ("mega encoding", True), model.HintRole)
        self.assertEqual(ind.data(), ("b", "mega encoding", True))
        self.assertEqual(ind.data(model.HintRole), ("mega encoding", True))
        self.assertIn("b", ind.data(model.FilterRole))
        self.assertIn("mega encoding", ind.data(model.FilterRole))
        self.assertNotIn("bmega encoding", ind.data(model.FilterRole))
        self.assertIsNone(ind.data(Qt.ToolTipRole))

        ind = model.index(3, 0)  # separator
        self.assertIsNone(ind.data())
        self.assertIsNone(ind.data(model.HintRole))
        self.assertIsNone(ind.data(model.FilterRole))

    def test_defaultcontmodel(self):
        model = DefaultContModel()
        self.assertEqual(1, model.rowCount(QModelIndex()))
        self.assertEqual(1, model.columnCount(QModelIndex()))
        ind = model.index(0, 0)
        spy = QSignalSpy(model.dataChanged)
        model.setMethod("mega encoding")
        self.assertEqual(spy[0][0].row(), 0)
        self.assertEqual(ind.data(), "Preset: mega encoding")
        self.assertIsNotNone(ind.data(Qt.DecorationRole))
        self.assertIsNotNone(ind.data(Qt.ToolTipRole))


class TestListViewDelegate(unittest.TestCase):
    def test_displaytext(self):
        delegate = ListViewSearch.Delegate()
        self.assertEqual(delegate.displayText(("a", "foo", False), Mock()),
                         "a")
        self.assertEqual(delegate.displayText(("a", "foo", True), Mock()),
                         "a: foo")
        delegate.set_default_hints(True)
        self.assertEqual(delegate.displayText(("a", "foo", False), Mock()),
                         "a: foo")
        delegate.set_default_hints(False)
        self.assertEqual(delegate.displayText(("a", "foo", False), Mock()),
                         "a")

    @patch.object(SeparatedListDelegate, "initStyleOption")
    def test_bold(self, _):
        delegate = ListViewSearch.Delegate()
        option = Mock()
        index = Mock()
        index.data = lambda role: ("foo", True) \
            if role == ContDomainModel.HintRole else None
        delegate.initStyleOption(option, index)
        option.font.setBold.assert_called_with(True)
        index.data = lambda role: ("foo", False) \
            if role == ContDomainModel.HintRole else None
        delegate.initStyleOption(option, index)
        option.font.setBold.assert_called_with(False)
        index.data = lambda role: None \
            if role == ContDomainModel.HintRole else None
        delegate.initStyleOption(option, index)
        option.font.setBold.assert_called_with(False)


if __name__ == "__main__":
    unittest.main()



"""

"""
import os
import sysconfig

import pkg_resources

from orangecanvas.registry import CategoryDescription
from orangecanvas.registry.utils import category_from_package_globals
import orangewidget.workflow.discovery


# Entry point for main Orange categories/widgets discovery
def widget_discovery(discovery):
    # type: (orangewidget.workflow.discovery.WidgetDiscovery) -> None
    dist = pkg_resources.get_distribution("Orange3")
    pkgs = [
        "Orange.widgets.data",
        "Orange.widgets.visualize",
        "Orange.widgets.model",
        "Orange.widgets.evaluate",
        "Orange.widgets.unsupervised",
    ]
    for pkg in pkgs:
        discovery.handle_category(category_from_package_globals(pkg))
    # manually described category (without 'package' definition)
    discovery.handle_category(
        CategoryDescription(
            name="Transform",
            priority=1,
            background="#FF9D5E",
            icon="data/icons/Transform.svg",
            package=__package__,
        )
    )
    discovery.handle_category(
        CategoryDescription(
            name="Orange Obsolete",
            package=__package__,
            hidden=True,
        )
    )
    for pkg in pkgs:
        discovery.process_category_package(pkg, distribution=dist)
    discovery.process_widget_module("Orange.widgets.obsolete.owtable")


WIDGET_HELP_PATH = (
    ("{DEVELOP_ROOT}/doc/visual-programming/build/htmlhelp/index.html", None),
    (os.path.join(sysconfig.get_path("data"),
                  "share/help/en/orange3/htmlhelp/index.html"),
     None),
    ("https://docs.biolab.si/orange/3/visual-programming/", ""),
)



# pylint: disable=protected-access
import unittest
from unittest.mock import patch

from AnyQt.QtCore import Qt

from orangewidget.tests.utils import excepthook_catch

from Orange.widgets.data.owtable import OWTable
from Orange.widgets.tests.base import WidgetTest, WidgetOutputsTestMixin
from Orange.data import Table, Domain
from Orange.data.sql.table import SqlTable
from Orange.tests.sql.base import DataBaseTest as dbt


class TestOWTable(WidgetTest, WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        WidgetOutputsTestMixin.init(cls,
                                    output_all_on_no_selection=True)

        cls.signal_name = OWTable.Inputs.data
        cls.signal_data = cls.data  # pylint: disable=no-member

    def setUp(self):
        super().setUp()
        self.widget = self.create_widget(OWTable)

    def test_input_data(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertIs(self.widget.input.table, self.data)
        self.assertIs(self.widget.view.model().source, self.data)
        self.send_signal(self.widget.Inputs.data, None)
        self.assertIsNone(self.widget.input)
        self.assertIsNone(self.widget.view.model())

    def test_input_data_empty(self):
        self.send_signal(self.widget.Inputs.data, self.data[:0])
        output = self.get_output(self.widget.Outputs.annotated_data)
        self.assertIsInstance(output, Table)
        self.assertEqual(len(output), 0)

    def test_data_model(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertEqual(self.widget.view.model().rowCount(), len(self.data))

    def test_reset_select(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self._select_data()
        self.send_signal(self.widget.Inputs.data, Table('heart_disease'))
        self.assertListEqual([], self.widget.stored_selection["columns"])
        self.assertListEqual([], self.widget.stored_selection["rows"])

    def _select_data(self):
        self.widget.set_selection(
            list(range(0, len(self.data), 10)),
            list(range(len(self.data.domain.variables))),
        )
        return self.widget.stored_selection["rows"]

    def test_attrs_appear_in_corner_text(self):
        domain = self.data.domain
        new_domain = Domain(
            domain.attributes[1:], domain.class_var, domain.attributes[:1])
        new_domain.metas[0].attributes = {"c": "foo"}
        new_domain.attributes[0].attributes = {"a": "bar", "c": "baz"}
        new_domain.class_var.attributes = {"b": "foo"}
        self.send_signal(self.widget.Inputs.data, self.data.transform(new_domain))
        self.assertEqual(self.widget.view.cornerText(), "\na\nb\nc")

    def test_unconditional_commit_on_new_signal(self):
        with patch.object(self.widget.commit, 'now') as commit:
            self.widget.auto_commit = False
            commit.reset_mock()
            self.send_signal(self.widget.Inputs.data, self.data)
            commit.assert_called()

    def test_pending_selection(self):
        widget = self.create_widget(OWTable, stored_settings={
            "stored_selection": {
                "rows": [5, 6, 7, 8, 9],
                "columns": list(range(len(self.data.domain.variables)))
            }
        })
        self.send_signal(widget.Inputs.data, None)
        self.send_signal(widget.Inputs.data, self.data)
        output = self.get_output(widget.Outputs.selected_data)
        self.assertEqual(5, len(output))

    def test_pending_sorted_selection(self):
        rows = [5, 6, 7, 8, 9, 55, 56, 57, 58, 59]
        widget = self.create_widget(OWTable, stored_settings={
            "stored_selection": {
                "rows": rows,
                "columns": list(range(len(self.data.domain.variables)))
            },
            "stored_sort": [("sepal length", 1), ("sepal width", -1)]
        })
        self.send_signal(widget.Inputs.data, None)
        self.send_signal(widget.Inputs.data, self.data)
        self.assertEqual(widget.view.horizontalHeader().sortIndicatorOrder(),
                         Qt.DescendingOrder)
        self.assertEqual(widget.view.horizontalHeader().sortIndicatorSection(), 2)
        output = self.get_output(widget.Outputs.selected_data)
        self.assertEqual(len(rows), len(output))
        sepal_width = output.get_column("sepal width").tolist()
        sepal_length = output.get_column("sepal length").tolist()
        self.assertSequenceEqual(sepal_width, sorted(sepal_width, reverse=True))
        dd = list(zip(sepal_length, sepal_width))
        dd_sorted = sorted(dd, key=lambda t: t[0])
        dd_sorted = sorted(dd_sorted, key=lambda t: t[1], reverse=True)
        self.assertSequenceEqual(dd, dd_sorted)
        ids = self.data[rows].ids
        self.assertSetEqual(set(output.ids), set(ids))

    def test_missing_sort_column_shows_warning(self):
        widget = self.create_widget(OWTable, stored_settings={
            "stored_sort": [("sepal length", 1), ("no such column", -1)]
        })
        self.send_signal(widget.Inputs.data, self.data)
        self.assertTrue(widget.Warning.missing_sort_columns.is_shown())
        self.send_signal(widget.Inputs.data, None)
        self.assertFalse(widget.Warning.missing_sort_columns.is_shown())

    def test_sorting(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.set_selection(
            [0, 1, 2, 3, 4],
            list(range(len(self.data.domain.variables)))
        )
        output = self.get_output(self.widget.Outputs.selected_data)
        output = output.get_column(0)
        output_original = output.tolist()

        self.widget.view.sortByColumn(1, Qt.AscendingOrder)
        self.assertEqual(self.widget.stored_sort, [('sepal length', 1)])
        output = self.get_output(self.widget.Outputs.selected_data)
        output = output.get_column(0)
        output_sorted = output.tolist()

        # the two outputs should not be the same.
        self.assertTrue(output_original != output_sorted)

        # check if output after sorting is actually sorted.
        self.assertTrue(sorted(output_original) == output_sorted)
        self.assertTrue(sorted(output_sorted) == output_sorted)

        self.widget.restore_order()
        output = self.get_output(self.widget.Outputs.selected_data)
        self.assertEqual(output.get_column(0).tolist(), output_original)

        # Check that output is the same with no sorting and cleared selection.
        self.widget.set_selection([], [])
        output = self.get_output(self.widget.Outputs.selected_data)
        self.assertIs(output, self.data)

    def test_sort_basket_column(self):
        data = self.data.to_sparse()
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.view.sortByColumn(0, Qt.AscendingOrder)
        self.assertEqual(self.widget.stored_sort, [("iris", 1)])
        self.widget.view.sortByColumn(1, Qt.AscendingOrder)
        self.assertEqual(self.widget.stored_sort,
                         [("iris", 1), ("\\BASKET(FEATURES)", 1)])
        # test restore
        w = self.create_widget(OWTable, stored_settings={
            "stored_sort": self.widget.stored_sort
        })
        self.send_signal(w.Inputs.data, data)
        self.assertEqual(w.stored_sort, self.widget.stored_sort)
        output_a = self.get_output(self.widget.Outputs.selected_data)
        output_b = self.get_output(w.Outputs.selected_data)
        self.assertEqual(output_a.ids.tolist(), output_b.ids.tolist())

    def test_info(self):
        info_text = self.widget.info_text
        no_input = "No data."
        self.assertEqual(info_text.text(), no_input)

    def test_show_distributions(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        # run through the delegate paint routines
        with excepthook_catch():
            w.grab()
        w.controls.show_distributions.toggle()
        with excepthook_catch():
            w.grab()
        w.controls.color_by_class.toggle()
        with excepthook_catch():
            w.grab()
        w.controls.show_distributions.toggle()

    def test_whole_rows(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        self.assertTrue(w.select_rows)  # default value
        with excepthook_catch():
            w.controls.select_rows.toggle()
        self.assertFalse(w.select_rows)
        w.set_selection([0, 1, 2, 3], [0, 1])
        out = self.get_output(w.Outputs.selected_data)
        self.assertEqual(out.domain,
                         Domain([self.data.domain.attributes[0]], self.data.domain.class_var))
        with excepthook_catch():
            w.controls.select_rows.toggle()
        out = self.get_output(w.Outputs.selected_data)
        self.assertTrue(w.select_rows)
        self.assertEqual(out.domain, self.data.domain)

    def test_show_attribute_labels(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        self.assertTrue(w.show_attribute_labels)  # default value
        with excepthook_catch():
            w.controls.show_attribute_labels.toggle()
        self.assertFalse(w.show_attribute_labels)

    def test_subset_input(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        with patch.object(w.signalManager, "send") as m:
            self.send_signal(w.Inputs.data_subset, self.data[[0, 1, 5]])
            m.assert_not_called()
        w.view.grab()  # cover delegate painting methods

        model = w.view.model()
        self.assertTrue(model.index(0, 0).data(model.SubsetRole))
        self.assertFalse(model.index(2, 0).data(model.SubsetRole))
        self.assertTrue(model.headerData(0, Qt.Vertical, model.SubsetRole))
        self.assertFalse(model.headerData(2, Qt.Vertical, model.SubsetRole))

        with patch.object(w.signalManager, "send") as m:
            self.send_signal(w.Inputs.data_subset, None)
            m.assert_not_called()

        w.view.grab()

        model = w.view.model()
        self.assertFalse(model.index(0, 0).data(model.SubsetRole))
        self.assertFalse(model.headerData(0, Qt.Vertical, model.SubsetRole))


class TestOWTableSQL(TestOWTable, dbt):
    def setUpDB(self):
        # pylint: disable=attribute-defined-outside-init
        conn, iris = self.create_iris_sql_table()
        data = SqlTable(conn, iris, inspect_values=True)
        if self.current_db == "mssql":
            # when loading data from mssql db, Sql widget returns Table (not SqlTable)
            data = Table(data)
        self.data = data.transform(Domain(data.domain.attributes[:-1],
                                          data.domain.attributes[-1]))

    def tearDownDB(self):
        self.drop_iris_sql_table()

    @dbt.run_on(["postgres", "mssql"])
    def test_input_data(self):
        super().test_input_data()

    @unittest.skip("no data output")
    def test_input_data_empty(self):
        super().test_input_data_empty()

    @unittest.skip("approx_len messes up row count")
    def test_data_model(self):
        super().test_data_model()

    @dbt.run_on(["postgres", "mssql"])
    def test_unconditional_commit_on_new_signal(self):
        super().test_unconditional_commit_on_new_signal()

    @dbt.run_on(["postgres", "mssql"])
    def test_reset_select(self):
        super().test_reset_select()

    @dbt.run_on(["postgres", "mssql"])
    def test_attrs_appear_in_corner_text(self):
        super().test_attrs_appear_in_corner_text()

    @unittest.skip("no data output")
    def test_pending_selection(self):
        super().test_pending_selection()

    @unittest.skip("sorting not implemented")
    def test_sorting(self):
        super().test_sorting()

    @unittest.skip("does nothing")
    def test_info(self):
        super().test_info()

    @dbt.run_on(["postgres", "mssql"])
    def test_show_distributions(self):
        super().test_show_distributions()

    @unittest.skip("no data output")
    def test_whole_rows(self):
        super().test_whole_rows()

    @dbt.run_on(["postgres", "mssql"])
    def test_show_attribute_labels(self):
        super().test_show_distributions()


if __name__ == "__main__":
    unittest.main()



import os.path
import unittest
from tempfile import TemporaryDirectory

import numpy as np

from Orange.data import (
    ContinuousVariable,
    guess_data_type,
    Table,
    Domain,
    StringVariable,
    DiscreteVariable,
)
from Orange.data.io_util import update_origin


class TestIoUtil(unittest.TestCase):
    def test_guess_continuous_w_nans(self):
        self.assertIs(
            guess_data_type(["9", "", "98", "?", "98", "98", "98"])[2],
            ContinuousVariable)


class TestUpdateOrigin(unittest.TestCase):
    FILE_NAMES = ["file1.txt", "file2.txt", "file3.txt"]

    def setUp(self) -> None:
        self.alt_dir = TemporaryDirectory()  # pylint: disable=consider-using-with

        self.var_string = var = StringVariable("Files")
        files = self.FILE_NAMES + [var.Unknown]
        self.table_string = Table.from_list(
            Domain([], metas=[var]), np.array(files).reshape((-1, 1))
        )
        self.var_discrete = var = DiscreteVariable("Files", values=self.FILE_NAMES)
        files = self.FILE_NAMES + [var.Unknown]
        self.table_discrete = Table.from_list(
            Domain([], metas=[var]), np.array(files).reshape((-1, 1))
        )

    def tearDown(self) -> None:
        self.alt_dir.cleanup()

    def __create_files(self):
        for f in self.FILE_NAMES:
            f = os.path.join(self.alt_dir.name, f)
            with open(f, "w", encoding="utf8"):
                pass
            self.assertTrue(os.path.exists(f))

    def test_origin_not_changed(self):
        """
        Origin exist; keep it unchanged, even though dataset path also includes
        files from column.
        """
        with TemporaryDirectory() as dir_name:
            self.var_string.attributes["origin"] = dir_name
            update_origin(self.table_string, self.alt_dir.name)
            self.assertEqual(
                self.table_string.domain[self.var_string].attributes["origin"], dir_name
            )

    def test_origin_subdir(self):
        """
        Origin is wrong but last dir in origin exit in the dataset file's path
        """
        images_dir = os.path.join(self.alt_dir.name, "subdir")
        os.mkdir(images_dir)

        self.var_string.attributes["origin"] = "/a/b/subdir"
        update_origin(self.table_string, os.path.join(self.alt_dir.name, "data.csv"))
        self.assertEqual(
            self.table_string.domain[self.var_string].attributes["origin"], images_dir
        )

    def test_origin_parents_subdir(self):
        """
        Origin is wrong but last dir in origin exit in the dataset file
        parent's directory
        """
        # make the dir where dataset is placed
        images_dir = os.path.join(self.alt_dir.name, "subdir")
        os.mkdir(images_dir)

        self.var_string.attributes["origin"] = "/a/b/subdir"
        update_origin(self.table_string, os.path.join(images_dir, "data.csv"))
        self.assertEqual(
            self.table_string.domain[self.var_string].attributes["origin"], images_dir
        )

    def test_column_paths_subdir(self):
        """
        Origin dir not exiting but paths from column exist in dataset's dir
        """
        self.__create_files()

        self.var_string.attributes["origin"] = "/a/b/non-exiting-dir"
        update_origin(self.table_string, os.path.join(self.alt_dir.name, "data.csv"))
        self.assertEqual(
            self.table_string.domain[self.var_string].attributes["origin"],
            self.alt_dir.name,
        )

        self.var_discrete.attributes["origin"] = "/a/b/non-exiting-dir"
        update_origin(self.table_discrete, os.path.join(self.alt_dir.name, "data.csv"))
        self.assertEqual(
            self.table_discrete.domain[self.var_discrete].attributes["origin"],
            self.alt_dir.name,
        )

    def test_column_paths_parents_subdir(self):
        """
        Origin dir not exiting but paths from column exist in dataset parent's dir
        """
        # make the dir where dataset is placed
        dataset_dir = os.path.join(self.alt_dir.name, "subdir")
        self.__create_files()

        self.var_string.attributes["origin"] = "/a/b/non-exiting-dir"
        update_origin(self.table_string, os.path.join(dataset_dir, "data.csv"))
        self.assertEqual(
            self.table_string.domain[self.var_string].attributes["origin"],
            self.alt_dir.name,
        )

        self.var_discrete.attributes["origin"] = "/a/b/non-exiting-dir"
        update_origin(self.table_discrete, os.path.join(dataset_dir, "data.csv"))
        self.assertEqual(
            self.table_discrete.domain[self.var_discrete].attributes["origin"],
            self.alt_dir.name,
        )


if __name__ == '__main__':
    unittest.main()



import os

import numpy as np

from AnyQt.QtWidgets import QSizePolicy, QStyle, QMessageBox, QFileDialog
from AnyQt.QtCore import QTimer

from orangewidget.settings import Setting
from orangewidget.widget import Msg
from orangewidget.workflow.drophandler import SingleFileDropHandler

from Orange.misc import DistMatrix
from Orange.widgets import widget, gui
from Orange.data import get_sample_datasets_dir
from Orange.widgets.utils.filedialogs import RecentPathsWComboMixin, RecentPath, \
    stored_recent_paths_prepend
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Output


class OWDistanceFile(widget.OWWidget, RecentPathsWComboMixin):
    name = "Distance File"
    id = "orange.widgets.unsupervised.distancefile"
    description = "Read distances from a file."
    icon = "icons/DistanceFile.svg"
    priority = 10
    keywords = "distance file, load, read, open"

    class Outputs:
        distances = Output("Distances", DistMatrix, dynamic=False)

    class Error(widget.OWWidget.Error):
        invalid_file = Msg("Data was not loaded:{}")
        non_square_matrix = Msg(
            "Matrix is not square. "
            "Reformat the file and use the File widget to read it.")

    want_main_area = False
    resizing_enabled = False

    auto_symmetric = Setting(True)

    def __init__(self):
        super().__init__()
        RecentPathsWComboMixin.__init__(self)
        self.distances = None

        vbox = gui.vBox(self.controlArea, "Distance File")
        box = gui.hBox(vbox)
        self.file_combo.setMinimumWidth(300)
        box.layout().addWidget(self.file_combo)
        self.file_combo.activated[int].connect(self.select_file)

        button = gui.button(box, self, '...', callback=self.browse_file)
        button.setIcon(self.style().standardIcon(QStyle.SP_DirOpenIcon))
        button.setSizePolicy(
            QSizePolicy.Maximum, QSizePolicy.Fixed)

        button = gui.button(
            box, self, "Reload", callback=self.reload, default=True)
        button.setIcon(self.style().standardIcon(QStyle.SP_BrowserReload))
        button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        vbox = gui.vBox(self.controlArea, "Options")
        gui.checkBox(
            vbox, self, "auto_symmetric",
            "Treat triangular matrices as symmetric",
            tooltip="If matrix is triangular, this will copy the data to the "
                    "other triangle",
            callback=self.commit
        )

        gui.rubber(self.buttonsArea)
        gui.button(
            self.buttonsArea, self, "Browse documentation datasets",
            callback=lambda: self.browse_file(True), autoDefault=False)
        gui.rubber(self.buttonsArea)

        self.set_file_list()
        QTimer.singleShot(0, self.open_file)

    def set_file_list(self):
        super().set_file_list()

    def reload(self):
        return self.open_file()

    def select_file(self, n):
        super().select_file(n)
        self.set_file_list()
        self.open_file()

    def browse_file(self, in_demos=False):
        if in_demos:
            start_file = get_sample_datasets_dir()
            if not os.path.exists(start_file):
                QMessageBox.information(
                    None, "File",
                    "Cannot find the directory with documentation datasets")
                return
        else:
            start_file = self.last_path() or os.path.expanduser("~/")

        filename, _ = QFileDialog.getOpenFileName(
            self, 'Open Distance File', start_file,
            "Excel File (*.xlsx);;Distance File (*.dst)")
        if not filename:
            return
        self.add_path(filename)
        self.open_file()

    def open_file(self):
        self.Error.clear()
        self.distances = None
        fn = self.last_path()
        if fn and not os.path.exists(fn):
            dir_name, basename = os.path.split(fn)
            if os.path.exists(os.path.join(".", basename)):
                fn = os.path.join(".", basename)
        if fn and fn != "(none)":
            try:
                distances = DistMatrix.from_file(fn)
            except Exception as exc:
                err = str(exc)
                self.Error.invalid_file(" \n"[len(err) > 40] + err)
            else:
                if distances.shape[0] != distances.shape[1]:
                    self.Error.non_square_matrix()
                else:
                    np.nan_to_num(distances)
                    self.distances = distances
                    _, filename = os.path.split(fn)
                    self.distances.name, _ = os.path.splitext(filename)
        self.commit()

    def commit(self):
        distances = self.distances
        if distances is not None:
            if self.auto_symmetric:
                distances = distances.auto_symmetricized()
            if np.any(np.isnan(distances)):
                distances = np.nan_to_num(distances)
        self.Outputs.distances.send(distances)

    def send_report(self):
        if not self.distances:
            self.report_paragraph("No data was loaded.")
        else:
            self.report_items([("File name", self.distances.name)])


class OWDistanceFileDropHandler(SingleFileDropHandler):
    WIDGET = OWDistanceFile

    def parametersFromFile(self, path):
        r = RecentPath(os.path.abspath(path), None, None,
                       os.path.basename(path))
        return {"recent_paths": stored_recent_paths_prepend(self.WIDGET, r)}

    def canDropFile(self, path: str) -> bool:
        return os.path.splitext(path)[1].lower() in (".dst", ".xlsx")


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWDistanceFile).run()



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest
import numpy as np

from Orange.data import Table
from Orange.preprocess import Continuize
from Orange.projection import FreeViz


class TestFreeviz(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table("iris")
        cls.housing = Table("housing")
        cls.zoo = Table("zoo")

    def test_basic(self):
        table = self.iris.copy()
        with table.unlocked():
            table[3, 3] = np.nan
        freeviz = FreeViz()
        model = freeviz(table)
        proj = model(table)
        self.assertEqual(len(proj), len(table))
        self.assertTrue(np.isnan(proj.X).any())
        np.testing.assert_array_equal(proj[:100], model(table[:100]))

    def test_regression(self):
        table = Table("housing")[::10]
        freeviz = FreeViz()
        freeviz(table)

        freeviz = FreeViz(p=2)
        freeviz(table)

    @unittest.skip("Test weights is too slow.")
    def test_weights(self):
        table = Table("iris")
        weights = np.random.rand(150, 1).flatten()
        freeviz = FreeViz(weights=weights, p=3, scale=False, center=False)
        freeviz(table)

        scale = np.array([0.5, 0.4, 0.6, 0.8])
        freeviz = FreeViz(scale=scale, center=[0.2, 0.6, 0.4, 0.2])
        freeviz(table)

    def test_raising_errors(self):
        table = Table("iris")

        scale = np.array([0.5, 0.4, 0.6])
        freeviz = FreeViz(scale=scale)
        self.assertRaises(ValueError, freeviz, table)

        freeviz = FreeViz(center=[0.6, 0.4, 0.2])
        self.assertRaises(ValueError, freeviz, table)

        weights = np.random.rand(100, 1).flatten()
        freeviz = FreeViz(weights=weights)
        self.assertRaises(ValueError, freeviz, table)

        table = Table("titanic")[::10]
        freeviz = FreeViz()
        self.assertRaises(ValueError, freeviz, table)

    def test_initial(self):
        FreeViz.init_radial(1)
        FreeViz.init_radial(2)
        FreeViz.init_radial(3)
        FreeViz.init_random(2, 4, 5)

    def test_transform_changed_domain(self):
        """
        1. Open data, apply some preprocessor, splits the data into two parts,
        use FreeViz on the first part, and then transform the second part.

        2. Open data, split into two parts, apply the same preprocessor and
        FreeViz only on the first part, and then transform the second part.

        The transformed second part in (1) and (2) has to be the same.
        """
        data = Table("titanic")[::10]
        normalize = Continuize()
        freeviz = FreeViz(maxiter=40)

        # normalize all
        ndata = normalize(data)
        model = freeviz(ndata[:100])
        result_1 = model(ndata[100:])

        # normalize only the "training" part
        ndata = normalize(data[:100])
        model = freeviz(ndata)
        result_2 = model(data[100:])

        np.testing.assert_almost_equal(result_1.X, result_2.X)



import sys
from itertools import chain, starmap
from typing import Sequence, Tuple, cast, Optional

import numpy as np

from AnyQt.QtCore import (
    Qt, QObject, QEvent, QSize, QAbstractProxyModel, QItemSelection,
    QItemSelectionModel, QItemSelectionRange, QAbstractItemModel
)
from AnyQt.QtGui import QPainter
from AnyQt.QtWidgets import (
    QStyle, QWidget, QStyleOptionHeader, QAbstractButton
)

import Orange.data
import Orange.data.sql.table

from Orange.widgets.data.utils.models import RichTableModel
from Orange.widgets.utils.itemmodels import TableModel
from Orange.widgets.utils.itemselectionmodel import (
    BlockSelectionModel, selection_blocks, ranges
)
from Orange.widgets.utils.tableview import TableView


class DataTableView(TableView):
    """
    A TableView with settable corner text.
    """
    class __CornerPainter(QObject):
        def drawCorner(self, button: QWidget):
            opt = QStyleOptionHeader()
            view = self.parent()
            assert isinstance(view, DataTableView)
            header = view.horizontalHeader()
            opt.initFrom(header)
            state = QStyle.State_None
            if button.isEnabled():
                state |= QStyle.State_Enabled
            if button.isActiveWindow():
                state |= QStyle.State_Active
            if button.isDown():
                state |= QStyle.State_Sunken
            opt.state = state
            opt.rect = button.rect()
            opt.text = button.text()
            opt.position = QStyleOptionHeader.OnlyOneSection
            style = header.style()
            painter = QPainter(button)
            style.drawControl(QStyle.CE_Header, opt, painter, header)

        def eventFilter(self, receiver: QObject, event: QEvent) -> bool:
            if event.type() == QEvent.Paint:
                self.drawCorner(receiver)
                return True
            return super().eventFilter(receiver, event)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__cornerText = ""
        self.__cornerButton = btn = self.findChild(QAbstractButton)
        self.__cornerButtonFilter = DataTableView.__CornerPainter(self)
        btn.installEventFilter(self.__cornerButtonFilter)
        btn.clicked.disconnect(self.selectAll)
        btn.clicked.connect(self.cornerButtonClicked)
        if sys.platform == "darwin":
            btn.setAttribute(Qt.WA_MacSmallSize)

    def setCornerText(self, text: str) -> None:
        """Set the corner text."""
        self.__cornerButton.setText(text)
        self.__cornerText = text
        self.__cornerButton.update()
        assert self.__cornerButton is self.findChild(QAbstractButton)
        opt = QStyleOptionHeader()
        opt.initFrom(self.__cornerButton)
        opt.text = text
        s = self.__cornerButton.style().sizeFromContents(
            QStyle.CT_HeaderSection, opt, QSize(), self.__cornerButton
        )
        if s.isValid():
            self.verticalHeader().setMinimumWidth(s.width())

    def cornerText(self):
        """Return the corner text."""
        return self.__cornerText

    def cornerButtonClicked(self):
        model = self.model()
        selection = self.selectionModel()
        selection = selection.selection()
        if len(selection) == 1:
            srange = selection[0]
            if srange.top() == 0 and srange.left() == 0 \
                    and srange.right() == model.columnCount() - 1 \
                    and srange.bottom() == model.rowCount() - 1:
                self.clearSelection()
            else:
                self.selectAll()
        else:
            self.selectAll()


def source_model(model: QAbstractItemModel) -> Optional[QAbstractItemModel]:
    while isinstance(model, QAbstractProxyModel):
        model = model.sourceModel()
    return model


def is_table_sortable(table):
    if isinstance(table, Orange.data.sql.table.SqlTable):
        return False
    elif isinstance(table, Orange.data.Table):
        return True
    else:
        return False


class RichTableView(DataTableView):
    """
    The preferred table view for RichTableModel.

    Handles the display of variable's labels keys in top left corner.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        header = self.horizontalHeader()
        header.setSortIndicator(-1, Qt.AscendingOrder)

    def setModel(self, model: QAbstractItemModel):
        current = self.model()
        if current is not None:
            current.headerDataChanged.disconnect(self.__headerDataChanged)
        super().setModel(model)
        if model is not None:
            model.headerDataChanged.connect(self.__headerDataChanged)
            self.__headerDataChanged(Qt.Horizontal)
            select_rows = self.selectionBehavior() == TableView.SelectRows
            sel_model = BlockSelectionModel(model, selectBlocks=not select_rows)
            self.setSelectionModel(sel_model)

            sortable = self.isModelSortable(model)
            self.setSortingEnabled(sortable)
            header = self.horizontalHeader()
            header.setSectionsClickable(sortable)
            header.setSortIndicatorShown(sortable)

    def isModelSortable(self, model: QAbstractItemModel) -> bool:
        """
        Should the `model` be sortable via the view header click.

        This predicate is called when a model is set on the view and
        enables/disables the model sorting and header section sort indicators.
        """
        model = source_model(model)
        if isinstance(model, TableModel):
            table = model.source
            return is_table_sortable(table)
        return False

    def __headerDataChanged(
            self,
            orientation: Qt.Orientation,
    ) -> None:
        if orientation == Qt.Horizontal:
            model = self.model()
            model = source_model(model)
            if isinstance(model, RichTableModel) and \
                    model.richHeaderFlags() & RichTableModel.Labels:
                items = model.headerData(
                    0, Qt.Horizontal, RichTableModel.LabelsItemsRole
                )
                text = "\n"
                text += "\n".join(key for key, _ in items)
            else:
                text = ""
            self.setCornerText(text)

    def setBlockSelection(
            self, rows: Sequence[int], columns: Sequence[int]
    ) -> None:
        """
        Set the block row and column selection.

        Note
        ----
        The `rows` indices refer to the underlying TableModel's rows.

        Parameters
        ----------
        rows: Sequence[int]
            The rows to select.
        columns: Sequence[int]
            The columns to select.

        See Also
        --------
        blockSelection()
        """
        model = self.model()
        if model is None:
            return
        sel_model = self.selectionModel()
        assert isinstance(sel_model, BlockSelectionModel)
        if not rows or not columns or model.rowCount() <= rows[-1] or \
                model.columnCount() <= columns[-1]:
            # selection out of range for the model
            rows = columns = []
        proxy_chain = []
        while isinstance(model, QAbstractProxyModel):
            proxy_chain.append(model)
            model = model.sourceModel()
        assert isinstance(model, TableModel)

        rows = model.mapFromSourceRows(rows)

        selection = QItemSelection()
        rowranges = list(ranges(rows))
        colranges = list(ranges(columns))

        for rowstart, rowend in rowranges:
            for colstart, colend in colranges:
                selection.append(
                    QItemSelectionRange(
                        model.index(rowstart, colstart),
                        model.index(rowend - 1, colend - 1)
                    )
                )
        for proxy in proxy_chain[::-1]:
            selection = proxy.mapSelectionFromSource(selection)
        sel_model.select(selection, QItemSelectionModel.ClearAndSelect)

    def blockSelection(self) -> Tuple[Sequence[int], Sequence[int]]:
        """
        Return the current selected rows and columns.

        Note
        ----
        The `rows` indices refer to the underlying TableModel's rows.
        """
        model = self.model()
        if model is None:
            return [], []
        sel_model = self.selectionModel()
        selection = sel_model.selection()

        # map through the proxies into input table.
        while isinstance(model, QAbstractProxyModel):
            selection = model.mapSelectionToSource(selection)
            model = model.sourceModel()

        assert isinstance(sel_model, BlockSelectionModel)
        assert isinstance(model, TableModel)

        row_spans, col_spans = selection_blocks(selection)
        rows = list(chain.from_iterable(starmap(range, row_spans)))
        cols = list(chain.from_iterable(starmap(range, col_spans)))
        rows = np.array(rows, dtype=np.intp)
        # map the rows through the applied sorting (if any)
        rows = model.mapToSourceRows(rows)
        rows = cast(Sequence[int], rows.tolist())
        return rows, cols



from collections import namedtuple
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, List, Optional, Set

import pandas as pd
from numpy import nan
from AnyQt.QtCore import (
    QAbstractTableModel,
    QEvent,
    QItemSelectionModel,
    QModelIndex,
    Qt,
)
from AnyQt.QtWidgets import (
    QAbstractItemView,
    QCheckBox,
    QGridLayout,
    QHeaderView,
    QTableView,
)
from orangewidget.settings import ContextSetting, Setting
from orangewidget.utils.listview import ListViewFilter
from orangewidget.utils.signals import Input, Output
from orangewidget.utils import enum_as_int
from orangewidget.widget import Msg
from pandas.core.dtypes.common import is_datetime64_any_dtype

from Orange.data import (
    ContinuousVariable,
    DiscreteVariable,
    Domain,
    StringVariable,
    Table,
    TimeVariable,
    Variable,
)
from Orange.data.aggregate import OrangeTableGroupBy
from Orange.util import wrap_callback
from Orange.widgets import gui
from Orange.widgets.settings import DomainContextHandler
from Orange.widgets.utils.concurrent import ConcurrentWidgetMixin, TaskState
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import OWWidget

Aggregation = namedtuple("Aggregation", ["function", "types"])


def concatenate(x):
    """
    Concatenate values of series if value is not missing (nan or empty string
    for StringVariable)
    """
    return " ".join(str(v) for v in x if not pd.isnull(v) and len(str(v)) > 0)


def std(s):
    """
    Std that also handle time variable. Pandas's std return Timedelta object in
    case of datetime columns - transform TimeDelta to seconds
    """
    std_ = s.std()
    if isinstance(std_, pd.Timedelta):
        return std_.total_seconds()
    # std returns NaT when cannot compute value - change it to nan to keep colum numeric
    return nan if pd.isna(std_) else std_


def var(s):
    """
    Variance that also handle time variable. Pandas's variance function somehow
    doesn't support DateTimeArray - this function fist converts datetime series
    to UNIX epoch and then computes variance
    """
    if is_datetime64_any_dtype(s):
        initial_ts = pd.Timestamp("1970-01-01", tz=None if s.dt.tz is None else "UTC")
        s = (s - initial_ts) / pd.Timedelta("1s")
    var_ = s.var()
    return var_.total_seconds() if isinstance(var_, pd.Timedelta) else var_


def span(s):
    """
    Span that also handle time variable. Time substitution return Timedelta
    object in case of datetime columns - transform TimeDelta to seconds
    """
    span_ = pd.Series.max(s) - pd.Series.min(s)
    return span_.total_seconds() if isinstance(span_, pd.Timedelta) else span_


AGGREGATIONS = {
    "Mean": Aggregation("mean", {ContinuousVariable, TimeVariable}),
    "Median": Aggregation("median", {ContinuousVariable, TimeVariable}),
    "Q1": Aggregation(lambda s: s.quantile(0.25), {ContinuousVariable, TimeVariable}),
    "Q3": Aggregation(lambda s: s.quantile(0.75), {ContinuousVariable, TimeVariable}),
    "Min. value": Aggregation("min", {ContinuousVariable, TimeVariable}),
    "Max. value": Aggregation("max", {ContinuousVariable, TimeVariable}),
    "Mode": Aggregation(
        lambda x: pd.Series.mode(x).get(0, nan),
        {ContinuousVariable, DiscreteVariable, TimeVariable}
    ),
    "Standard deviation": Aggregation(std, {ContinuousVariable, TimeVariable}),
    "Variance": Aggregation(var, {ContinuousVariable, TimeVariable}),
    "Sum": Aggregation("sum", {ContinuousVariable}),
    "Concatenate": Aggregation(
        concatenate,
        {ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable},
    ),
    "Span": Aggregation(span, {ContinuousVariable, TimeVariable}),
    "First value": Aggregation(
        "first", {ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable}
    ),
    "Last value": Aggregation(
        "last", {ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable}
    ),
    "Random value": Aggregation(
        lambda x: x.sample(1, random_state=0),
        {ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable},
    ),
    "Count defined": Aggregation(
        "count", {ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable}
    ),
    "Count": Aggregation(
        "size", {ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable}
    ),
    "Proportion defined": Aggregation(
        lambda x: x.count() / x.size,
        {ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable},
    ),
}
# list of ordered aggregation names is required on several locations so we
# prepare it in advance
AGGREGATIONS_ORD = list(AGGREGATIONS)

# use first aggregation suitable for each type as default
DEFAULT_AGGREGATIONS = {
    var: {next(name for name, agg in AGGREGATIONS.items() if var in agg.types)}
    for var in (ContinuousVariable, TimeVariable, DiscreteVariable, StringVariable)
}


@dataclass
class Result:
    group_by: OrangeTableGroupBy = None
    result_table: Optional[Table] = None


def _run(
    data: Table,
    group_by_attrs: List[Variable],
    aggregations: Dict[Variable, Set[str]],
    result: Result,
    state: TaskState,
) -> Result:
    def progress(part):
        state.set_progress_value(part * 100)
        if state.is_interruption_requested():
            raise Exception

    state.set_status("Aggregating")
    # group table rows
    if result.group_by is None:
        result.group_by = data.groupby(group_by_attrs)
    state.set_partial_result(result)

    aggregations = {
        var: [
            (agg, AGGREGATIONS[agg].function)
            for agg in sorted(aggs, key=AGGREGATIONS_ORD.index)
        ]
        for var, aggs in aggregations.items()
    }
    result.result_table = result.group_by.aggregate(
        aggregations, wrap_callback(progress, 0.2, 1)
    )
    return result


class TabColumn:
    attribute = 0
    aggregations = 1


TABLE_COLUMN_NAMES = ["Attributes", "Aggregations"]


class VarTableModel(QAbstractTableModel):
    def __init__(self, parent: "OWGroupBy", *args):
        super().__init__(*args)
        self.domain = None
        self.parent = parent

    def set_domain(self, domain: Domain) -> None:
        """
        Reset the table view to new domain
        """
        self.domain = domain
        self.modelReset.emit()

    def update_aggregation(self, attribute: str) -> None:
        """
        Reset the aggregation values in the table for the attribute
        """
        index = self.domain.index(attribute)
        if index < 0:
            # indices of metas are negative: first meta -1, second meta -2, ...
            index = len(self.domain.variables) - 1 - index
        index = self.index(index, 1)
        self.dataChanged.emit(index, index)

    def rowCount(self, parent=None) -> int:
        return (
            0
            if self.domain is None or (parent is not None and parent.isValid())
            else len(self.domain.variables) + len(self.domain.metas)
        )

    @staticmethod
    def columnCount(parent=None) -> int:
        return 0 if parent is not None and parent.isValid() else len(TABLE_COLUMN_NAMES)

    def data(self, index, role=Qt.DisplayRole) -> Any:
        row, col = index.row(), index.column()
        val = (self.domain.variables + self.domain.metas)[row]
        if role in (Qt.DisplayRole, Qt.EditRole):
            if col == TabColumn.attribute:
                return str(val)
            else:  # col == TabColumn.aggregations
                # plot first two aggregations comma separated and write n more
                # for others
                aggs = sorted(
                    self.parent.aggregations.get(val, []), key=AGGREGATIONS_ORD.index
                )
                n_more = "" if len(aggs) <= 3 else f" and {len(aggs) - 3} more"
                return ", ".join(aggs[:3]) + n_more
        elif role == Qt.DecorationRole and col == TabColumn.attribute:
            return gui.attributeIconDict[val]
        return None

    def headerData(self, i, orientation, role=Qt.DisplayRole) -> str:
        if orientation == Qt.Horizontal and role == Qt.DisplayRole and i < 2:
            return TABLE_COLUMN_NAMES[i]
        return super().headerData(i, orientation, role)


class AggregateListViewSearch(ListViewFilter):
    """ListViewSearch that disables unselecting all items in the list"""

    def selectionCommand(
        self, index: QModelIndex, event: QEvent = None
    ) -> QItemSelectionModel.SelectionFlags:
        flags = super().selectionCommand(index, event)
        selmodel = self.selectionModel()
        if not index.isValid():  # Click on empty viewport; don't clear
            return QItemSelectionModel.NoUpdate
        if selmodel.isSelected(index):
            currsel = selmodel.selectedIndexes()
            if len(currsel) == 1 and index == currsel[0]:
                # Is the last selected index; do not deselect it
                return QItemSelectionModel.NoUpdate
        if (
            event is not None
            and event.type() == QEvent.MouseMove
            and flags & QItemSelectionModel.ToggleCurrent
        ):
            # Disable ctrl drag 'toggle'; can be made to deselect the last
            # index, would need to keep track of the current selection
            # (selectionModel does this but does not expose it)
            flags &= ~QItemSelectionModel.Toggle
            flags |= QItemSelectionModel.Select
        return flags


class CheckBox(QCheckBox):
    def __init__(self, text, parent):
        super().__init__(text)
        self.parent: OWGroupBy = parent

    def nextCheckState(self) -> None:
        """
        Custom behaviour for switching between steps. It is required since
        sometimes user will select different types of attributes at the same
        time. In this case we step between unchecked, partially checked and
        checked or just between unchecked and checked - depending on situation
        """
        if self.checkState() == Qt.Checked:
            # if checked always uncheck
            self.setCheckState(Qt.Unchecked)
        else:
            agg = self.text()
            selected_attrs = self.parent.get_selected_attributes()
            types = set(type(attr) for attr in selected_attrs)
            can_be_applied_all = types <= AGGREGATIONS[agg].types

            # true if aggregation applied to all attributes that can be
            # aggregated with selected aggregation
            applied_all = all(
                type(attr) not in AGGREGATIONS[agg].types
                or agg in self.parent.aggregations[attr]
                for attr in selected_attrs
            )
            if self.checkState() == Qt.PartiallyChecked:
                # if partially check: 1) check if agg can be applied to all
                # 2) else uncheck if agg already applied to all
                # 3) else leve partially checked to apply to all that can be aggregated
                if can_be_applied_all:
                    self.setCheckState(Qt.Checked)
                elif applied_all:
                    self.setCheckState(Qt.Unchecked)
                else:
                    self.setCheckState(Qt.PartiallyChecked)
                    # since checkbox state stay same signal is not emitted
                    # automatically but we need a callback call so we emit it
                    self.stateChanged.emit(enum_as_int(Qt.PartiallyChecked))
            else:  # self.checkState() == Qt.Unchecked
                # if unchecked: check if all can be checked else partially check
                self.setCheckState(
                    Qt.Checked if can_be_applied_all else Qt.PartiallyChecked
                )


@contextmanager
def block_signals(widget):
    widget.blockSignals(True)
    try:
        yield
    finally:
        widget.blockSignals(False)


class OWGroupBy(OWWidget, ConcurrentWidgetMixin):
    name = "Group by"
    description = ""
    category = "Transform"
    icon = "icons/GroupBy.svg"
    keywords = "aggregate, group by"
    priority = 1210

    class Inputs:
        data = Input("Data", Table, doc="Input data table")

    class Outputs:
        data = Output("Data", Table, doc="Aggregated data")

    class Error(OWWidget.Error):
        unexpected_error = Msg("{}")

    settingsHandler = DomainContextHandler()

    gb_attrs: List[Variable] = ContextSetting([])
    aggregations: Dict[Variable, Set[str]] = ContextSetting({})
    auto_commit: bool = Setting(True)

    def __init__(self):
        super().__init__()
        ConcurrentWidgetMixin.__init__(self)

        self.data = None
        self.result = None

        self.gb_attrs_model = DomainModel(
            separators=False,
        )
        self.agg_table_model = VarTableModel(self)
        self.agg_checkboxes = {}

        self.__init_control_area()
        self.__init_main_area()

    def __init_control_area(self) -> None:
        """Init all controls in the control area"""
        gui.listView(
            self.controlArea,
            self,
            "gb_attrs",
            box="Group by",
            model=self.gb_attrs_model,
            viewType=AggregateListViewSearch,
            callback=self.__gb_changed,
            selectionMode=ListViewFilter.ExtendedSelection,
        )

        gui.auto_send(self.buttonsArea, self, "auto_commit")

    def __init_main_area(self) -> None:
        """Init all controls in the main area"""
        # aggregation table
        self.agg_table_view = tableview = QTableView()
        tableview.setModel(self.agg_table_model)
        tableview.setSelectionBehavior(QAbstractItemView.SelectRows)
        tableview.selectionModel().selectionChanged.connect(self.__rows_selected)
        tableview.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        vbox = gui.vBox(self.mainArea, " ")
        vbox.layout().addWidget(tableview)

        # aggregations checkboxes
        grid_layout = QGridLayout()
        gui.widgetBox(self.mainArea, orientation=grid_layout, box="Aggregations")

        col = 0
        row = 0
        break_rows = (6, 6, 99)
        for agg in AGGREGATIONS:
            self.agg_checkboxes[agg] = cb = CheckBox(agg, self)
            cb.setDisabled(True)
            cb.stateChanged.connect(partial(self.__aggregation_changed, agg))
            grid_layout.addWidget(cb, row, col)
            row += 1
            if row == break_rows[col]:
                row = 0
                col += 1

    ############
    # Callbacks

    def __rows_selected(self) -> None:
        """Callback for table selection change; update checkboxes"""
        selected_attrs = self.get_selected_attributes()

        types = {type(attr) for attr in selected_attrs}
        active_aggregations = [self.aggregations[attr] for attr in selected_attrs]
        for agg, cb in self.agg_checkboxes.items():
            cb.setDisabled(not types & AGGREGATIONS[agg].types)

            activated = {agg in a for a in active_aggregations}
            with block_signals(cb):
                # check if aggregation active for all selected attributes,
                # partially check if active for some else uncheck
                cb.setCheckState(
                    Qt.Checked
                    if activated == {True}
                    else (Qt.Unchecked if activated == {False} else Qt.PartiallyChecked)
                )

    def __gb_changed(self) -> None:
        """Callback for Group-by attributes selection change"""
        self.result = Result()
        self.commit.deferred()

    def __aggregation_changed(self, agg: str) -> None:
        """
        Callback for aggregation change; update aggregations dictionary and call
        commit
        """
        selected_attrs = self.get_selected_attributes()
        for attr in selected_attrs:
            if self.agg_checkboxes[agg].isChecked() and self.__aggregation_compatible(
                agg, attr
            ):
                self.aggregations[attr].add(agg)
            else:
                self.aggregations[attr].discard(agg)
            self.agg_table_model.update_aggregation(attr)
        self.commit.deferred()

    @Inputs.data
    def set_data(self, data: Table) -> None:
        self.closeContext()
        self.data = data

        # reset states
        self.cancel()
        self.result = Result()
        self.Outputs.data.send(None)
        self.gb_attrs_model.set_domain(data.domain if data else None)
        self.gb_attrs = self.gb_attrs_model[:1] if self.gb_attrs_model else []
        self.aggregations = (
            {
                attr: DEFAULT_AGGREGATIONS[type(attr)].copy()
                for attr in data.domain.variables + data.domain.metas
            }
            if data
            else {}
        )
        default_aggregations = self.aggregations.copy()

        self.openContext(self.data)

        # restore aggregations
        self.aggregations.update({k: v for k, v in default_aggregations.items()
                                  if k not in self.aggregations})

        # update selections in widgets and re-plot
        self.agg_table_model.set_domain(data.domain if data else None)
        self._set_gb_selection()

        self.commit.now()

    #########################
    # Task connected methods

    @gui.deferred
    def commit(self) -> None:
        self.Error.clear()
        self.Warning.clear()
        if self.data:
            self.start(_run, self.data, self.gb_attrs, self.aggregations, self.result)

    def on_done(self, result: Result) -> None:
        self.result = result
        self.Outputs.data.send(result.result_table)

    def on_partial_result(self, result: Result) -> None:
        # store result in case the task is canceled and on_done is not called
        self.result = result

    def on_exception(self, ex: Exception):
        self.Error.unexpected_error(str(ex))

    ###################
    # Helper methods

    def get_selected_attributes(self):
        """Get select attributes in the table"""
        selection_model = self.agg_table_view.selectionModel()
        sel_rows = selection_model.selectedRows()
        vars_ = self.data.domain.variables + self.data.domain.metas
        return [vars_[index.row()] for index in sel_rows]

    def _set_gb_selection(self) -> None:
        """
        Update selected attributes. When context includes variable hidden in
        data, it will match and gb_attrs may include hidden attribute. Remove it
        since otherwise widget groups by attribute that is not present in view.
        """
        values = self.gb_attrs_model[:]
        self.gb_attrs = [var_ for var_ in self.gb_attrs if var_ in values]
        if not self.gb_attrs and self.gb_attrs_model:
            # if gb_attrs empty select first
            self.gb_attrs = self.gb_attrs_model[:1]

    @staticmethod
    def __aggregation_compatible(agg, attr):
        """Check a compatibility of aggregation with the variable"""
        return type(attr) in AGGREGATIONS[agg].types

    @classmethod
    def migrate_context(cls, context, _):
        """
        Before widget allowed using Sum on Time variable, now it is forbidden.
        This function removes Sum from the context for TimeVariables (104)
        """
        for var_, v in context.values["aggregations"][0].items():
            if len(var_) == 2:
                if var_[1] == 104:
                    v.discard("Sum")


if __name__ == "__main__":
    # pylint: disable=ungrouped-imports
    from orangewidget.utils.widgetpreview import WidgetPreview

    WidgetPreview(OWGroupBy).run(Table("iris"))



"""
Adapted from a code editor component created
for Enki editor as replacement for QScintilla.
Copyright (C) 2020  Andrei Kopats

Originally licensed under the terms of GNU Lesser General Public License
as published by the Free Software Foundation, version 2.1 of the license.
This is compatible with Orange3's GPL-3.0 license.
"""  # pylint: disable=duplicate-code
import unittest

from Orange.widgets.data.utils.pythoneditor.tests.base import EditorTest


class Test(EditorTest):
    def _ws_test(self,
                 text,
                 expectedResult,
                 drawAny=None,
                 drawIncorrect=None,
                 useTab=None,
                 indentWidth=None):
        if drawAny is None:
            drawAny = [True, False]
        if drawIncorrect is None:
            drawIncorrect = [True, False]
        if useTab is None:
            useTab = [True, False]
        if indentWidth is None:
            indentWidth = [1, 2, 3, 4, 8]
        for drawAnyVal in drawAny:
            self.qpart.drawAnyWhitespace = drawAnyVal

            for drawIncorrectVal in drawIncorrect:
                self.qpart.drawIncorrectIndentation = drawIncorrectVal

                for useTabVal in useTab:
                    self.qpart.indentUseTabs = useTabVal

                    for indentWidthVal in indentWidth:
                        self.qpart.indentWidth = indentWidthVal
                        try:
                            self._verify(text, expectedResult)
                        except:
                            print("Failed params:\n\tany {}\n\tincorrect {}\n\ttabs {}\n\twidth {}"
                                  .format(self.qpart.drawAnyWhitespace,
                                          self.qpart.drawIncorrectIndentation,
                                          self.qpart.indentUseTabs,
                                          self.qpart.indentWidth))
                            raise

    def _verify(self, text, expectedResult):
        res = self.qpart._chooseVisibleWhitespace(text)  # pylint: disable=protected-access
        for index, value in enumerate(expectedResult):
            if value == '1':
                if not res[index]:
                    self.fail("Item {} is not True:\n\t{}".format(index, res))
            elif value == '0':
                if res[index]:
                    self.fail("Item {} is not False:\n\t{}".format(index, res))
            else:
                assert value == ' '

    def test_1(self):
        # Trailing
        self._ws_test('   m xyz\t ',
                      '   0 00011',
                      drawIncorrect=[True])

    def test_2(self):
        # Tabs in space mode
        self._ws_test('\txyz\t',
                      '10001',
                      drawIncorrect=[True], useTab=[False])

    def test_3(self):
        # Spaces in tab mode
        self._ws_test('    2   3     5',
                      '111100000000000',
                      drawIncorrect=[True], drawAny=[False], indentWidth=[3], useTab=[True])

    def test_4(self):
        # Draw any
        self._ws_test(' 1 1  2   3     5\t',
                      '100011011101111101',
                      drawAny=[True],
                      indentWidth=[2, 3, 4, 8])


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from tempfile import NamedTemporaryFile

import numpy as np

from Orange.data import ContinuousVariable, DiscreteVariable, StringVariable, \
    TimeVariable, Domain, Table
from Orange.data.io import TabReader, ExcelReader
from Orange.data.io_util import guess_data_type
from Orange.misc.collections import natural_sorted


class TestTableFilters(unittest.TestCase):
    def test_guess_data_type_continuous(self):
        # should be ContinuousVariable
        valuemap, values, coltype = guess_data_type(list(range(1, 100)))
        self.assertEqual(ContinuousVariable, coltype)
        self.assertIsNone(valuemap)
        np.testing.assert_array_equal(np.array(list(range(1, 100))), values)

        valuemap, values, coltype = guess_data_type([1, 2, 3, 1, 2, 3])
        self.assertEqual(ContinuousVariable, coltype)
        self.assertIsNone(valuemap)
        np.testing.assert_array_equal([1, 2, 3, 1, 2, 3], values)

        valuemap, values, coltype = guess_data_type(
            ["1", "2", "3", "1", "2", "3"])
        self.assertEqual(ContinuousVariable, coltype)
        self.assertIsNone(valuemap)
        np.testing.assert_array_equal([1, 2, 3, 1, 2, 3], values)

    def test_guess_data_type_discrete(self):
        # should be DiscreteVariable
        valuemap, values, coltype = guess_data_type([1, 2, 1, 2])
        self.assertEqual(DiscreteVariable, coltype)
        self.assertEqual([1, 2], valuemap)
        np.testing.assert_array_equal([1, 2, 1, 2], values)

        valuemap, values, coltype = guess_data_type(["1", "2", "1", "2", "a"])
        self.assertEqual(DiscreteVariable, coltype)
        self.assertEqual(["1", "2", "a"], valuemap)
        np.testing.assert_array_equal(['1', '2', '1', '2', 'a'], values)

        # just below the threshold for string variable
        in_values = list(map(lambda x: str(x) + "a", range(24))) + ["a"] * 76
        valuemap, values, coltype = guess_data_type(in_values)
        self.assertEqual(DiscreteVariable, coltype)
        self.assertEqual(natural_sorted(set(in_values)), valuemap)
        np.testing.assert_array_equal(in_values, values)

    def test_guess_data_type_string(self):
        # should be StringVariable
        # too many different values for discrete
        in_values = list(map(lambda x: str(x) + "a", range(90)))
        valuemap, values, coltype = guess_data_type(in_values)
        self.assertEqual(StringVariable, coltype)
        self.assertIsNone(valuemap)
        np.testing.assert_array_equal(in_values, values)

        # more than len(values)**0.7
        in_values = list(map(lambda x: str(x) + "a", range(25))) + ["a"] * 75
        valuemap, values, coltype = guess_data_type(in_values)
        self.assertEqual(StringVariable, coltype)
        self.assertIsNone(valuemap)
        np.testing.assert_array_equal(in_values, values)

        # more than 100 different values - exactly 101
        # this is the case when len(values)**0.7 rule would vote for the
        # DiscreteVariable
        in_values = list(map(lambda x: str(x) + "a", range(100))) + ["a"] * 999
        valuemap, values, coltype = guess_data_type(in_values)
        self.assertEqual(StringVariable, coltype)
        self.assertIsNone(valuemap)
        np.testing.assert_array_equal(in_values, values)

    def test_guess_data_type_time(self):
        in_values = ["2019-10-10", "2019-10-10", "2019-10-10", "2019-10-01"]
        valuemap, _, coltype = guess_data_type(in_values)
        self.assertEqual(TimeVariable, coltype)
        self.assertIsNone(valuemap)

        in_values = ["2019-10-10T12:08:51", "2019-10-10T12:08:51",
                     "2019-10-10T12:08:51", "2019-10-01T12:08:51"]
        valuemap, _, coltype = guess_data_type(in_values)
        self.assertEqual(TimeVariable, coltype)
        self.assertIsNone(valuemap)

        in_values = ["2019-10-10 12:08:51", "2019-10-10 12:08:51",
                     "2019-10-10 12:08:51", "2019-10-01 12:08:51"]
        valuemap, _, coltype = guess_data_type(in_values)
        self.assertEqual(TimeVariable, coltype)
        self.assertIsNone(valuemap)

        in_values = ["2019-10-10 12:08", "2019-10-10 12:08",
                     "2019-10-10 12:08", "2019-10-01 12:08"]
        valuemap, _, coltype = guess_data_type(in_values)
        self.assertEqual(TimeVariable, coltype)
        self.assertIsNone(valuemap)

    def test_guess_data_type_values_order(self):
        """
        Test if values are ordered naturally
        """
        in_values = [
            "something1", "something12", "something2", "something1",
            "something20", "something1", "something2", "something12",
            "something1", "something12"
        ]
        res = ["something1", "something2", "something12", "something20"]
        valuemap, _, coltype = guess_data_type(in_values)
        self.assertEqual(DiscreteVariable, coltype)
        self.assertListEqual(res, valuemap)

class TestWriters(unittest.TestCase):
    def setUp(self):
        self.domain = Domain([DiscreteVariable("a", values=tuple("xyz")),
                         ContinuousVariable("b", number_of_decimals=3)],
                        ContinuousVariable("c", number_of_decimals=0),
                        [StringVariable("d")])
        self.data = Table.from_numpy(
            self.domain,
            np.array([[1, 0.5], [2, np.nan], [np.nan, 1.0625]]),
            np.array([3, 1, 7]),
            np.array(["foo bar baz".split()]).T
        )

    def test_write_tab(self):
        with NamedTemporaryFile(suffix=".tab", delete=False) as f:
            fname = f.name
        try:
            TabReader.write(fname, self.data)
            with open(fname, encoding="utf-8") as f:
                self.assertEqual(f.read().strip(), """
c\td\ta\tb
continuous\tstring\tx y z\tcontinuous
class\tmeta\t\t
3\tfoo\ty\t0.500
1\tbar\tz\t
7\tbaz\t\t1.06250""".strip())
        finally:
            os.remove(fname)

    def test_roundtrip_xlsx(self):
        with NamedTemporaryFile(suffix=".xlsx", delete=False) as f:
            fname = f.name
        try:
            ExcelReader.write(fname, self.data)
            data = ExcelReader(fname).read()
            np.testing.assert_equal(data.X, self.data.X)
            np.testing.assert_equal(data.Y, self.data.Y)
            np.testing.assert_equal(data.metas, self.data.metas)
            np.testing.assert_equal(data.domain, self.data.domain)
        finally:
            os.remove(fname)


if __name__ == "__main__":
    unittest.main()



from math import isnan

from AnyQt.QtCore import Qt, QIdentityProxyModel, QModelIndex

from orangewidget.gui import OrangeUserRole

import Orange
from Orange.widgets import gui
from Orange.widgets.utils.itemmodels import TableModel

_BarRole = gui.TableBarItem.BarRole


class RichTableModel(TableModel):
    """A TableModel with some extra bells and whistles/

    (adds support for gui.BarRole, include variable labels and icons
    in the header)
    """
    #: Rich header data flags.
    Name, Labels, Icon = 1, 2, 4

    #: Qt.ItemData role to retrieve variable's header attributes.
    LabelsItemsRole = next(OrangeUserRole)

    def __init__(self, sourcedata, parent=None):
        super().__init__(sourcedata, parent)

        self._header_flags = RichTableModel.Name
        self._continuous = [var.is_continuous for var in self.vars]
        labels = []
        for var in self.vars:
            if isinstance(var, Orange.data.Variable):
                labels.extend(var.attributes.keys())
        self._labels = list(sorted(
            {label for label in labels if not label.startswith("_")}))

    def data(self, index, role=Qt.DisplayRole):
        # pylint: disable=arguments-differ
        if role == _BarRole and self._continuous[index.column()]:
            val = super().data(index, TableModel.ValueRole)
            if val is None or isnan(val):
                return None

            dist = super().data(index, TableModel.VariableStatsRole)
            if dist is not None and dist.max > dist.min:
                return (val - dist.min) / (dist.max - dist.min)
            else:
                return None
        elif role == Qt.TextAlignmentRole and self._continuous[index.column()]:
            return Qt.AlignRight | Qt.AlignVCenter
        else:
            return super().data(index, role)

    def headerData(self, section, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            var = super().headerData(
                section, orientation, TableModel.VariableRole)
            if var is None:
                return super().headerData(
                    section, orientation, Qt.DisplayRole)

            lines = []
            if self._header_flags & RichTableModel.Name:
                lines.append(var.name)
            if self._header_flags & RichTableModel.Labels:
                lines.extend(str(var.attributes.get(label, ""))
                             for label in self._labels)
            return "\n".join(lines)
        elif orientation == Qt.Horizontal and \
                role == RichTableModel.LabelsItemsRole:
            var = super().headerData(
                section, orientation, TableModel.VariableRole)
            if var is None:
                return []
            return [(label, var.attributes.get(label))
                    for label in self._labels]
        elif orientation == Qt.Horizontal and role == Qt.DecorationRole and \
                self._header_flags & RichTableModel.Icon:
            var = super().headerData(
                section, orientation, TableModel.VariableRole)
            if var is not None:
                return gui.attributeIconDict[var]
            else:
                return None
        else:
            return super().headerData(section, orientation, role)

    def setRichHeaderFlags(self, flags):
        if flags != self._header_flags:
            self._header_flags = flags
            self.headerDataChanged.emit(
                Qt.Horizontal, 0, self.columnCount() - 1)

    def richHeaderFlags(self):
        return self._header_flags


# This is used for sub-setting large (SQL) models. Largely untested probably
# broken.
class TableSliceProxy(QIdentityProxyModel):
    def __init__(self, parent=None, rowSlice=slice(0, None, 1), **kwargs):
        super().__init__(parent, **kwargs)
        self.__rowslice = slice(0, None, 1)
        self.setRowSlice(rowSlice)

    def setRowSlice(self, rowslice):
        if rowslice.step is not None and rowslice.step != 1:
            raise ValueError("invalid stride")

        if self.__rowslice != rowslice:
            self.beginResetModel()
            self.__rowslice = rowslice
            self.endResetModel()

    def index(
            self, row: int, column: int, _parent: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        return self.createIndex(row, column)

    def parent(self, _child: QModelIndex) -> QModelIndex:
        return QModelIndex()

    def sibling(self, row: int, column: int, _idx: QModelIndex) -> QModelIndex:
        return self.index(row, column)

    def mapToSource(self, proxyindex):
        model = self.sourceModel()
        if model is None or not proxyindex.isValid():
            return QModelIndex()

        row, col = proxyindex.row(), proxyindex.column()
        row = row + self.__rowslice.start
        if 0 <= row < model.rowCount():
            return model.createIndex(row, col)
        else:
            return QModelIndex()

    def mapFromSource(self, sourceindex):
        model = self.sourceModel()
        if model is None or not sourceindex.isValid():
            return QModelIndex()
        row, col = sourceindex.row(), sourceindex.column()
        row = row - self.__rowslice.start
        if 0 <= row < self.rowCount():
            return self.createIndex(row, col)
        else:
            return QModelIndex()

    def rowCount(self, parent=QModelIndex()):
        if parent.isValid():
            return 0
        count = super().rowCount()
        start, stop, step = self.__rowslice.indices(count)
        assert step == 1
        return stop - start



import itertools
import concurrent.futures

from collections import namedtuple
from typing import List, Optional

import numpy
from scipy.sparse import issparse

from AnyQt.QtWidgets import QTableView, QHeaderView, QApplication, QStyle
from AnyQt.QtGui import QColor, QClipboard
from AnyQt.QtCore import (
    Qt, QSize, QMetaObject,
    QAbstractProxyModel,
    QItemSelectionModel, QItemSelection, QItemSelectionRange,
)
from AnyQt.QtCore import pyqtSlot as Slot

import Orange.data
from Orange.data.table import Table
from Orange.data.sql.table import SqlTable

from Orange.widgets import gui
from Orange.widgets.data.utils.tableview import RichTableView
from Orange.widgets.settings import Setting
from Orange.widgets.data.utils.models import TableSliceProxy, RichTableModel
from Orange.widgets.utils.itemdelegates import TableDataDelegate
from Orange.widgets.utils.itemselectionmodel import (
    BlockSelectionModel, ranges, selection_blocks
)
from Orange.widgets.utils.tableview import table_selection_to_mime_data
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, MultiInput, Output, Msg
from Orange.widgets.utils.annotated_data import (create_annotated_table,
                                                 ANNOTATED_DATA_SIGNAL_NAME)
from Orange.widgets.utils.itemmodels import TableModel
from Orange.widgets.utils.state_summary import format_summary_details
from Orange.widgets.data.utils import tablesummary as tsummary


TableSlot = namedtuple("TableSlot", ["input_id", "table", "summary", "view"])


class DataTableView(gui.HScrollStepMixin, RichTableView):
    dataset: Table
    input_slot: TableSlot


class TableBarItemDelegate(gui.TableBarItem, TableDataDelegate):
    pass


class OWDataTable(OWWidget):
    category = "Orange Obsolete"
    replaces = ["Orange.widgets.data.owtable.OWDataTable"]

    name = "Data Table"
    description = "View the dataset in a spreadsheet."
    icon = "../data/icons/Table.svg"
    priority = 50
    keywords = []

    class Inputs:
        data = MultiInput("Data", Table, auto_summary=False, filter_none=True)

    class Outputs:
        selected_data = Output("Selected Data", Table, default=True)
        annotated_data = Output(ANNOTATED_DATA_SIGNAL_NAME, Table)

    class Warning(OWWidget.Warning):
        multiple_inputs = Msg(
            "Multiple Data inputs are deprecated.\n"
            "This functionality will be removed soon.\n"
            "Use multiple Tables instead.")

    buttons_area_orientation = Qt.Vertical

    show_distributions = Setting(False)
    show_attribute_labels = Setting(True)
    select_rows = Setting(True)
    auto_commit = Setting(True)

    color_by_class = Setting(True)
    selected_rows = Setting([], schema_only=True)
    selected_cols = Setting([], schema_only=True)

    settings_version = 2

    def __init__(self):
        super().__init__()
        self._inputs: List[TableSlot] = []
        self.__pending_selected_rows = self.selected_rows
        self.selected_rows = None
        self.__pending_selected_cols = self.selected_cols
        self.selected_cols = None

        self.dist_color = QColor(220, 220, 220, 255)

        info_box = gui.vBox(self.controlArea, "Info")
        self.info_text = gui.widgetLabel(info_box)
        self._set_input_summary(None)

        box = gui.vBox(self.controlArea, "Variables")
        self.c_show_attribute_labels = gui.checkBox(
            box, self, "show_attribute_labels",
            "Show variable labels (if present)",
            callback=self._on_show_variable_labels_changed)

        gui.checkBox(box, self, "show_distributions",
                     'Visualize numeric values',
                     callback=self._on_distribution_color_changed)
        gui.checkBox(box, self, "color_by_class", 'Color by instance classes',
                     callback=self._on_distribution_color_changed)

        box = gui.vBox(self.controlArea, "Selection")

        gui.checkBox(box, self, "select_rows", "Select full rows",
                     callback=self._on_select_rows_changed)

        gui.rubber(self.controlArea)

        gui.button(self.buttonsArea, self, "Restore Original Order",
                   callback=self.restore_order,
                   tooltip="Show rows in the original order",
                   autoDefault=False,
                   attribute=Qt.WA_LayoutUsesWidgetRect)
        gui.auto_send(self.buttonsArea, self, "auto_commit")

        # GUI with tabs
        self.tabs = gui.tabWidget(self.mainArea)
        self.tabs.currentChanged.connect(self._on_current_tab_changed)

    def copy_to_clipboard(self):
        self.copy()

    def sizeHint(self):
        return QSize(800, 500)

    def _create_table_view(self):
        view = DataTableView()
        view.setSortingEnabled(True)
        view.setItemDelegate(TableDataDelegate(view))

        if self.select_rows:
            view.setSelectionBehavior(QTableView.SelectRows)

        header = view.horizontalHeader()
        header.setSectionsMovable(True)
        header.setSectionsClickable(True)
        header.setSortIndicatorShown(True)
        header.setSortIndicator(-1, Qt.AscendingOrder)

        # QHeaderView does not 'reset' the model sort column,
        # because there is no guaranty (requirement) that the
        # models understand the -1 sort column.
        def sort_reset(index, order):
            if view.model() is not None and index == -1:
                view.model().sort(index, order)
        header.sortIndicatorChanged.connect(sort_reset)
        return view

    @Inputs.data
    def set_dataset(self, index: int, data: Table):
        """Set the input dataset."""
        datasetname = getattr(data, "name", "Data")
        slot = self._inputs[index]
        view = slot.view
        # reset the (header) view state.
        view.setModel(None)
        view.horizontalHeader().setSortIndicator(-1, Qt.AscendingOrder)
        assert self.tabs.indexOf(view) != -1
        self.tabs.setTabText(self.tabs.indexOf(view), datasetname)
        view.dataset = data
        slot = TableSlot(index, data, tsummary.table_summary(data), view)
        view.input_slot = slot
        self._inputs[index] = slot
        self._setup_table_view(view, data)
        self.tabs.setCurrentWidget(view)
        self._set_multi_input_warning()

    @Inputs.data.insert
    def insert_dataset(self, index: int, data: Table):
        datasetname = getattr(data, "name", "Data")
        view = self._create_table_view()
        slot = TableSlot(None, data, tsummary.table_summary(data), view)
        view.dataset = data
        view.input_slot = slot
        self._inputs.insert(index, slot)
        self.tabs.insertTab(index, view, datasetname)
        self._setup_table_view(view, data)
        self.tabs.setCurrentWidget(view)
        self._set_multi_input_warning()

    @Inputs.data.remove
    def remove_dataset(self, index):
        slot = self._inputs.pop(index)
        view = slot.view
        self.tabs.removeTab(self.tabs.indexOf(view))
        view.setModel(None)
        view.hide()
        view.deleteLater()

        current = self.tabs.currentWidget()
        if current is not None:
            self._set_input_summary(current.input_slot)
        self._set_multi_input_warning()

    def _set_multi_input_warning(self):
        self.Warning.multiple_inputs(shown=len(self._inputs) > 1)

    def handleNewSignals(self):
        super().handleNewSignals()
        self.tabs.tabBar().setVisible(self.tabs.count() > 1)
        data: Optional[Table] = None
        current = self.tabs.currentWidget()
        slot = None
        if current is not None:
            data = current.dataset
            slot = current.input_slot

        if slot and isinstance(slot.summary.len, concurrent.futures.Future):
            def update(_):
                QMetaObject.invokeMethod(
                    self, "_update_info", Qt.QueuedConnection)
            slot.summary.len.add_done_callback(update)
        self._set_input_summary(slot)

        if data is not None and self.__pending_selected_rows is not None:
            self.selected_rows = self.__pending_selected_rows
            self.__pending_selected_rows = None
        else:
            self.selected_rows = []

        if data and self.__pending_selected_cols is not None:
            self.selected_cols = self.__pending_selected_cols
            self.__pending_selected_cols = None
        else:
            self.selected_cols = []

        self.set_selection()
        self.commit.now()

    def _setup_table_view(self, view, data):
        """Setup the `view` (QTableView) with `data` (Orange.data.Table)
        """
        datamodel = RichTableModel(data)
        rowcount = data.approx_len()

        if self.color_by_class and data.domain.has_discrete_class:
            color_schema = [
                QColor(*c) for c in data.domain.class_var.colors]
        else:
            color_schema = None
        if self.show_distributions:
            view.setItemDelegate(
                TableBarItemDelegate(
                    view, color=self.dist_color, color_schema=color_schema)
            )
        else:
            view.setItemDelegate(TableDataDelegate(view))

        header = view.horizontalHeader()
        header.sortIndicatorChanged.connect(self.update_selection)

        view.setModel(datamodel)

        vheader = view.verticalHeader()
        option = view.viewOptions()
        size = view.style().sizeFromContents(
            QStyle.CT_ItemViewItem, option,
            QSize(20, 20), view)

        vheader.setDefaultSectionSize(size.height() + 2)
        vheader.setMinimumSectionSize(5)
        vheader.setSectionResizeMode(QHeaderView.Fixed)

        # Limit the number of rows displayed in the QTableView
        # (workaround for QTBUG-18490 / QTBUG-28631)
        maxrows = (2 ** 31 - 1) // (vheader.defaultSectionSize() + 2)
        if rowcount > maxrows:
            sliceproxy = TableSliceProxy(
                parent=view, rowSlice=slice(0, maxrows))
            sliceproxy.setSourceModel(datamodel)
            # First reset the view (without this the header view retains
            # it's state - at this point invalid/broken)
            view.setModel(None)
            view.setModel(sliceproxy)

        assert view.model().rowCount() <= maxrows
        assert vheader.sectionSize(0) > 1 or datamodel.rowCount() == 0

        # update the header (attribute names)
        self._update_variable_labels(view)

        selmodel = BlockSelectionModel(
            view.model(), parent=view, selectBlocks=not self.select_rows)
        view.setSelectionModel(selmodel)
        view.selectionFinished.connect(self.update_selection)

    def _set_input_summary(self, slot):
        def format_summary(summary):
            if isinstance(summary, tsummary.ApproxSummary):
                length = summary.len.result() if summary.len.done() else \
                    summary.approx_len
            elif isinstance(summary, tsummary.Summary):
                length = summary.len
            return length

        summary, details = self.info.NoInput, ""
        if slot:
            summary = format_summary(slot.summary)
            details = format_summary_details(slot.table)
        self.info.set_input_summary(summary, details)
        if slot is None:
            summary = ["No data."]
        else:
            summary = tsummary.format_summary(slot.summary)
        self.info_text.setText("\n".join(summary))

    def _on_current_tab_changed(self, index):
        """Update the status bar on current tab change"""
        view = self.tabs.widget(index)
        if view is not None and view.model() is not None:
            self._set_input_summary(view.input_slot)
            self.update_selection()
        else:
            self._set_input_summary(None)

    def _update_variable_labels(self, view):
        "Update the variable labels visibility for `view`"
        model = view.model()
        if isinstance(model, TableSliceProxy):
            model = model.sourceModel()

        if self.show_attribute_labels:
            model.setRichHeaderFlags(
                RichTableModel.Labels | RichTableModel.Name)
        else:
            model.setRichHeaderFlags(RichTableModel.Name)

    def _on_show_variable_labels_changed(self):
        """The variable labels (var.attribues) visibility was changed."""
        for slot in self._inputs:
            self._update_variable_labels(slot.view)

    def _on_distribution_color_changed(self):
        for ti in range(self.tabs.count()):
            widget = self.tabs.widget(ti)
            model = widget.model()
            while isinstance(model, QAbstractProxyModel):
                model = model.sourceModel()
            data = model.source
            class_var = data.domain.class_var
            if self.color_by_class and class_var and class_var.is_discrete:
                color_schema = [QColor(*c) for c in class_var.colors]
            else:
                color_schema = None
            if self.show_distributions:
                delegate = TableBarItemDelegate(widget, color=self.dist_color,
                                                color_schema=color_schema)
            else:
                delegate = TableDataDelegate(widget)
            widget.setItemDelegate(delegate)
        tab = self.tabs.currentWidget()
        if tab:
            tab.reset()

    def _on_select_rows_changed(self):
        for slot in self._inputs:
            selection_model = slot.view.selectionModel()
            selection_model.setSelectBlocks(not self.select_rows)
            if self.select_rows:
                slot.view.setSelectionBehavior(QTableView.SelectRows)
                # Expand the current selection to full row selection.
                selection_model.select(
                    selection_model.selection(),
                    QItemSelectionModel.Select | QItemSelectionModel.Rows
                )
            else:
                slot.view.setSelectionBehavior(QTableView.SelectItems)

    def restore_order(self):
        """Restore the original data order of the current view."""
        table = self.tabs.currentWidget()
        if table is not None:
            table.horizontalHeader().setSortIndicator(-1, Qt.AscendingOrder)

    @Slot()
    def _update_info(self):
        current = self.tabs.currentWidget()
        if current is not None and current.model() is not None:
            self._set_input_summary(current.input_slot)

    def update_selection(self, *_):
        self.commit.deferred()

    def set_selection(self):
        if self.selected_rows and self.selected_cols:
            view = self.tabs.currentWidget()
            model = view.model()
            if model.rowCount() <= self.selected_rows[-1] or \
                    model.columnCount() <= self.selected_cols[-1]:
                return

            selection = QItemSelection()
            rowranges = list(ranges(self.selected_rows))
            colranges = list(ranges(self.selected_cols))

            for rowstart, rowend in rowranges:
                for colstart, colend in colranges:
                    selection.append(
                        QItemSelectionRange(
                            view.model().index(rowstart, colstart),
                            view.model().index(rowend - 1, colend - 1)
                        )
                    )
            view.selectionModel().select(
                selection, QItemSelectionModel.ClearAndSelect)

    @staticmethod
    def get_selection(view):
        """
        Return the selected row and column indices of the selection in view.
        """
        selmodel = view.selectionModel()

        selection = selmodel.selection()
        model = view.model()
        # map through the proxies into input table.
        while isinstance(model, QAbstractProxyModel):
            selection = model.mapSelectionToSource(selection)
            model = model.sourceModel()

        assert isinstance(selmodel, BlockSelectionModel)
        assert isinstance(model, TableModel)

        row_spans, col_spans = selection_blocks(selection)
        rows = list(itertools.chain.from_iterable(itertools.starmap(range, row_spans)))
        cols = list(itertools.chain.from_iterable(itertools.starmap(range, col_spans)))
        rows = numpy.array(rows, dtype=numpy.intp)
        # map the rows through the applied sorting (if any)
        rows = model.mapToSourceRows(rows)
        rows = rows.tolist()
        return rows, cols

    @staticmethod
    def _get_model(view):
        model = view.model()
        while isinstance(model, QAbstractProxyModel):
            model = model.sourceModel()
        return model

    @gui.deferred
    def commit(self):
        """
        Commit/send the current selected row/column selection.
        """
        selected_data = table = rowsel = None
        view = self.tabs.currentWidget()
        if view and view.model() is not None:
            model = self._get_model(view)
            table = model.source  # The input data table

            # Selections of individual instances are not implemented
            # for SqlTables
            if isinstance(table, SqlTable):
                self.Outputs.selected_data.send(selected_data)
                self.Outputs.annotated_data.send(None)
                return

            rowsel, colsel = self.get_selection(view)
            self.selected_rows, self.selected_cols = rowsel, colsel

            domain = table.domain

            if len(colsel) < len(domain.variables) + len(domain.metas):
                # only a subset of the columns is selected
                allvars = domain.class_vars + domain.metas + domain.attributes
                columns = [(c, model.headerData(c, Qt.Horizontal,
                                                TableModel.DomainRole))
                           for c in colsel]
                assert all(role is not None for _, role in columns)

                def select_vars(role):
                    """select variables for role (TableModel.DomainRole)"""
                    return [allvars[c] for c, r in columns if r == role]

                attrs = select_vars(TableModel.Attribute)
                if attrs and issparse(table.X):
                    # for sparse data you can only select all attributes
                    attrs = table.domain.attributes
                class_vars = select_vars(TableModel.ClassVar)
                metas = select_vars(TableModel.Meta)
                domain = Orange.data.Domain(attrs, class_vars, metas)

            # Send all data by default
            if not rowsel:
                selected_data = table
            else:
                selected_data = table.from_table(domain, table, rowsel)

        self.Outputs.selected_data.send(selected_data)
        self.Outputs.annotated_data.send(create_annotated_table(table, rowsel))

    def copy(self):
        """
        Copy current table selection to the clipboard.
        """
        view = self.tabs.currentWidget()
        if view is not None:
            mime = table_selection_to_mime_data(view)
            QApplication.clipboard().setMimeData(
                mime, QClipboard.Clipboard
            )

    def send_report(self):
        view = self.tabs.currentWidget()
        if not view or not view.model():
            return
        model = self._get_model(view)
        self.report_data_brief(model.source)
        self.report_table(view)


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWDataTable).run(
        insert_dataset=[
            (0, Table("iris")),
            (1, Table("brown-selected")),
            (2, Table("housing"))
        ])
